# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for serving serialized test cases for the reproduce tool."""

from flask import request

from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from handlers import base_handler
from libs import access
from libs import handler


def _prepare_testcase_dict(testcase):
  """Prepare a dictionary containing all information needed by the tool."""
  # By calling _to_dict directly here we prevent the need to modify this as
  # the testcase and other models changes over time.
  # pylint: disable=protected-access
  testcase_dict = testcase._to_dict()
  fuzz_target = data_handler.get_fuzz_target(testcase.actual_fuzzer_name())
  if fuzz_target:
    fuzz_target_dict = fuzz_target._to_dict()
  else:
    fuzz_target_dict = None
  # pylint: enable=protected-access

  # Several nonstandard bits of information are required for the tool to run.
  # Append these to the test case dict and serialize them as well.
  job = data_types.Job.query(data_types.Job.name == testcase.job_type).get()
  testcase_dict['job_definition'] = job.get_environment_string()
  testcase_dict['serialized_fuzz_target'] = fuzz_target_dict

  return testcase_dict


class Handler(base_handler.Handler):
  """Handler that returns a serialized testcase as JSON."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.oauth
  def post(self):
    """Serve the testcase JSON."""
    testcase_id = request.get('testcaseId')
    testcase = access.check_access_and_get_testcase(testcase_id)

    testcase_dict = _prepare_testcase_dict(testcase)
    return self.render_json(testcase_dict)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the severity analyzer module."""

import os
import unittest

from clusterfuzz._internal.crash_analysis import severity_analyzer
from clusterfuzz._internal.datastore.data_types import SecuritySeverity

DATA_DIRECTORY = os.path.join(
    os.path.dirname(__file__), 'stack_parsing', 'stack_analyzer_data')


class SeverityAnalyzerTest(unittest.TestCase):
  """Severity analyzer tests."""

  def _read_test_data(self, name):
    """Helper function to read test data."""
    with open(os.path.join(DATA_DIRECTORY, name)) as handle:
      return handle.read()

  def test_manual_severity_marker(self):
    """Tests that manual severity markers in uncaught exceptions are
    recognized."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Uncaught exception',
            self._read_test_data('java_severity_medium_exception.txt'), False),
        SecuritySeverity.MEDIUM)

  def test_asan_uaf(self):
    """Tests severity analysis of a use after free report."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Heap-use-after-free\nREAD 8', self._read_test_data('asan_uaf.txt'),
            False), SecuritySeverity.HIGH)

  def test_asan_uaf_gestures(self):
    """Tests severity analysis of a use after free report that requires
    gestures."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Heap-use-after-free\nREAD 8', self._read_test_data('asan_uaf.txt'),
            True), SecuritySeverity.MEDIUM)

  def test_asan_heap_overflow_read(self):
    """Tests severity analysis of a heap overflow read."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Heap-buffer-overflow\nREAD 4',
            self._read_test_data('asan_heap_overflow_read.txt'), False),
        SecuritySeverity.MEDIUM)

  def test_asan_heap_overflow_read_gestures(self):
    """Tests severity analysis of a heap overflow read that requires
    gestures."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Heap-buffer-overflow\nREAD 4',
            self._read_test_data('asan_heap_overflow_read.txt'), True),
        SecuritySeverity.LOW)

  def test_asan_heap_overflow_write(self):
    """Tests severity analysis of a heap overflow write."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Heap-buffer-overflow\nWRITE 16',
            self._read_test_data('asan_heap_overflow_write.txt'), False),
        SecuritySeverity.HIGH)

  def test_asan_heap_overflow_write_gestures(self):
    """Tests severity analysis of a heap overflow write that requires
    gestures."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Heap-buffer-overflow\nWRITE 16',
            self._read_test_data('asan_heap_overflow_write.txt'), True),
        SecuritySeverity.MEDIUM)

  def test_asan_container_overflow_read(self):
    """Tests severity analysis of a container overflow read."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Container-overflow\nREAD 4',
            self._read_test_data('asan_container_overflow_read.txt'), False),
        SecuritySeverity.MEDIUM)

  def test_asan_container_overflow_write(self):
    """Tests severity analysis of a container overflow write."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Container-overflow\nWRITE 4',
            self._read_test_data('asan_container_overflow_read.txt'), False),
        SecuritySeverity.HIGH)

  def test_sanitizer_chrome_renderer_uaf(self):
    """Tests severity analysis of a use after free in the renderer."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_chrome').analyze(
            'Heap-use-after-free\nREAD 8', self._read_test_data('asan_uaf.txt'),
            False), SecuritySeverity.HIGH)

  def test_asan_unknown_read(self):
    """Tests severity analysis of a SEGV read."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'UNKNOWN READ', self._read_test_data('asan_unknown_read.txt'),
            False), SecuritySeverity.MEDIUM)

  def test_asan_unknown_write(self):
    """Tests severity analysis of a SEGV write."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'UNKNOWN WRITE', self._read_test_data('asan_unknown_write.txt'),
            False), SecuritySeverity.HIGH)

  def test_tsan_uaf(self):
    """Tests severity analysis of a use after free report given by TSan."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Heap-use-after-free READ 1',
            self._read_test_data('tsan_use_after_free.txt'), False),
        SecuritySeverity.HIGH)

  def test_ubsan_bad_cast(self):
    """Tests severity analysis of a bad cast report given by UBSan vtpr."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Bad-cast', self._read_test_data('ubsan_bad_cast_downcast.txt'),
            False), SecuritySeverity.HIGH)

  def test_ubsan_incorrect_function_pointer_type(self):
    """Tests severity analysis of an incorrect function pointer type report
    given by UBSan function."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Incorrect-function-pointer-type',
            self._read_test_data('ubsan_incorrect_function_pointer_type.txt'),
            False), SecuritySeverity.MEDIUM)

  def test_ubsan_non_positive_vla_bound_value(self):
    """Tests severity analysis of a non-positive variable length array bound
    report given by UBSan function."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Non-positive-vla-bound-value',
            self._read_test_data('ubsan_non_positive_vla_bound_value.txt'),
            False), SecuritySeverity.MEDIUM)

  def test_ubsan_object_size(self):
    """Tests severity analysis of a object-size report given by UBSan
    function."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Object-size', self._read_test_data('ubsan_object_size.txt'),
            False), SecuritySeverity.MEDIUM)

  def test_cfi_bad_cast(self):
    """Tests severity analysis of a bad cast report given by CFI."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_generic').analyze(
            'Bad-cast', self._read_test_data('cfi_bad_cast.txt'), False),
        SecuritySeverity.HIGH)

  def test_msan_renderer(self):
    """Tests severity analysis of a MSan report in the renderer process."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_chrome').analyze(
            'Use-of-uninitialized-value',
            self._read_test_data('msan_renderer.txt'), False),
        SecuritySeverity.MEDIUM)

  def test_msan_browser(self):
    """Tests severity analysis of a MSan report in the browser process."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_chrome').analyze(
            'Use-of-uninitialized-value',
            self._read_test_data('msan_browser.txt'), False),
        SecuritySeverity.MEDIUM)

  def test_find_process_type_browser(self):
    """Tests a browser process bug is recognized as such."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_chrome').analyze(
            'Heap-use-after-free\nREAD 8',
            self._read_test_data('browser_uaf.txt'), False),
        SecuritySeverity.CRITICAL)

  def test_find_process_type_browser_2(self):
    """Tests a browser process bug is recognized as such (second variant)."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_chrome').analyze(
            'Heap-use-after-free\nREAD 8',
            self._read_test_data('browser_uaf2.txt'), False),
        SecuritySeverity.CRITICAL)

  def test_find_process_type_browser_3(self):
    """Tests a browser process bug is recognized as such (third variant)."""
    self.assertEqual(
        severity_analyzer.get_analyzer('sanitizer_chrome').analyze(
            'Heap-use-after-free\nREAD 8',
            self._read_test_data('browser_uaf3.txt'), False),
        SecuritySeverity.CRITICAL)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""build_manager tests."""
# pylint: disable=protected-access

import functools
import os
import shutil
import tempfile
import types
import unittest

import mock
import parameterized
from pyfakefs import fake_filesystem_unittest
import six

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.tasks import fuzz_task
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils

FAKE_APP_NAME = 'app'


# pylint: disable=unused-argument
def _mock_unpack_build(
    self,
    _,
    build_dir,
    build_url,
    target_weights=None,
):
  """Mock _unpack_build."""
  if not shell.remove_directory(build_dir, recreate=True):
    return False

  with open(os.path.join(build_dir, FAKE_APP_NAME), 'w') as f:
    f.write('')

  with open(os.path.join(build_dir, 'args.gn'), 'w') as f:
    f.write('')

  with open(os.path.join(build_dir, 'llvm-symbolizer'), 'w') as f:
    f.write('')

  fuzz_target = os.environ.get('FUZZ_TARGET')
  if fuzz_target:
    with open(os.path.join(build_dir, '.partial_build'), 'w') as f:
      f.write('')

  return True


def _get_timestamp(base_build_dir):
  """Return the timestamp, or None."""
  return utils.read_data_from_file(
      os.path.join(base_build_dir, '.timestamp'), eval_data=True)


class TrunkBuildTest(unittest.TestCase):
  """Tests for setting up trunk build."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager._setup_build_directories',
        'clusterfuzz._internal.build_management.build_manager.get_build_urls_list',
        'clusterfuzz._internal.build_management.build_manager.setup_regular_build',
    ])

    os.environ['BUILDS_DIR'] = '/builds'
    os.environ['RELEASE_BUILD_BUCKET_PATH'] = (
        'gs://path/file-release-([0-9]+).zip')
    os.environ['SYM_RELEASE_BUILD_BUCKET_PATH'] = (
        'gs://path/file-sym-release-([0-9]+).zip')
    os.environ['SYM_DEBUG_BUILD_BUCKET_PATH'] = (
        'gs://path/file-sym-debug-([0-9]+).zip')

  def test_setup_success(self):
    """Test successful setup."""
    self.mock.get_build_urls_list.side_effect = (
        [
            'gs://path/file-release-10.zip',
            'gs://path/file-release-2.zip',
            'gs://path/file-release-1.zip',
        ],
        [
            'gs://path/file-sym-release-10.zip',
            'gs://path/file-sym-release-2.zip',
            'gs://path/file-sym-release-1.zip',
        ],
        [
            'gs://path/file-sym-debug-10.zip',
            'gs://path/file-sym-debug-2.zip',
            'gs://path/file-sym-debug-1.zip',
        ],
    )

    build_manager.setup_build()
    self.mock.setup_regular_build.assert_called_with(
        10,
        'gs://path/file-release-([0-9]+).zip',
        build_prefix=None,
        target_weights=None)

  def test_setup_mismatch(self):
    """Test setup finding the first matching revision."""
    self.mock.get_build_urls_list.side_effect = (
        [
            'gs://path/file-release-10.zip',
            'gs://path/file-release-2.zip',
            'gs://path/file-release-1.zip',
        ],
        [
            'gs://path/file-sym-release-11.zip',
            'gs://path/file-sym-release-2.zip',
            'gs://path/file-sym-release-1.zip',
        ],
        [
            'gs://path/file-sym-debug-10.zip',
            'gs://path/file-sym-debug-2.zip',
            'gs://path/file-sym-debug-1.zip',
        ],
    )

    build_manager.setup_build()
    self.mock.setup_regular_build.assert_called_with(
        2,
        'gs://path/file-release-([0-9]+).zip',
        build_prefix=None,
        target_weights=None)

  def test_setup_fail(self):
    """Test setup failing to find any matching revisions."""
    self.mock.get_build_urls_list.side_effect = (
        [
            'gs://path/file-release-10.zip',
            'gs://path/file-release-3.zip',
            'gs://path/file-release-1.zip',
        ],
        [
            'gs://path/file-sym-release-11.zip',
            'gs://path/file-sym-release-2.zip',
            'gs://path/file-sym-release-1.zip',
        ],
        [
            'gs://path/file-sym-debug-10.zip',
            'gs://path/file-sym-debug-2.zip',
            'gs://path/file-sym-debug-0.zip',
        ],
    )

    build_manager.setup_build()
    self.assertEqual(0, self.mock.setup_regular_build.call_count)


@unittest.skipIf(
    not environment.get_value('FUCHSIA_TESTS'),
    'Temporarily disabling the Fuchsia test until build size reduced.')
class FuchsiaBuildTest(unittest.TestCase):
  """Tests for Fuchsia build setup."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.system.shell.clear_temp_directory',
    ])

    self.temp_dir = tempfile.mkdtemp()
    builds_dir = os.path.join(self.temp_dir, 'builds')
    os.mkdir(builds_dir)
    urls_dir = os.path.join(self.temp_dir, 'urls')
    os.mkdir(urls_dir)

    environment.set_value('JOB_NAME', 'libfuzzer_asan_fuchsia')
    environment.set_value('FAIL_RETRIES', 1)
    environment.set_value('BUILDS_DIR', builds_dir)
    environment.set_value('BUILD_URLS_DIR', urls_dir)
    environment.set_value('UNPACK_ALL_FUZZ_TARGETS_AND_FILES', True)
    environment.set_value(
        'RELEASE_BUILD_BUCKET_PATH',
        'gs://clusterfuchsia-builds-test/libfuzzer/'
        'fuchsia-([0-9]+).zip')
    environment.set_value('OS_OVERRIDE', 'FUCHSIA')

    self.maxDiff = None  # pylint: disable=invalid-name

  def tearDown(self):
    shutil.rmtree(self.temp_dir)

  def test_setup(self):
    """Tests setting up a build."""
    build = build_manager.setup_build(
        target_weights={'example_fuzzers/trap_fuzzer': 1000000.0})
    self.assertIsInstance(build, build_manager.FuchsiaBuild)

    self.assertIsNotNone(os.environ['FUZZ_TARGET'])
    self.assertEqual('example_fuzzers/trap_fuzzer', os.environ['FUZZ_TARGET'])
    self.assertEqual(20190926201257, environment.get_value('APP_REVISION'))

    # pylint: disable=protected-access
    targets = build._get_fuzz_targets_from_dir(build.build_dir)
    six.assertCountEqual(self, [
        'example_fuzzers/baz_fuzzer',
        'example_fuzzers/overflow_fuzzer',
        'example_fuzzers/trap_fuzzer',
        'ledger_fuzzers/p2p_sync_fuzzer',
        'ledger_fuzzers/encoding_fuzzer',
        'ledger_fuzzers/commit_pack_fuzzer',
        'bluetooth_fuzzers/basic_mode_rx_engine_fuzzer',
        'bluetooth_fuzzers/enhanced_retransmission_mode_rx_engine_fuzzer',
        'mdns_fuzzers/packet_reader_fuzzer',
        'zircon_fuzzers/nhlt-fuzzer',
        'zircon_fuzzers/zstd-fuzzer',
        'zircon_fuzzers/utf_conversion-fuzzer',
        'zircon_fuzzers/zbi-bootfs-fuzzer',
        'zircon_fuzzers/lz4-decompress-fuzzer',
        'zircon_fuzzers/lz4-fuzzer',
        'zircon_fuzzers/noop-fuzzer',
    ], targets)


class RegularBuildTest(fake_filesystem_unittest.TestCase):
  """Tests for regular build setup."""

  def setUp(self):
    """Setup for regular build test."""
    test_utils.set_up_pyfakefs(self)

    test_helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager.get_build_urls_list',
        'clusterfuzz._internal.build_management.build_manager.Build._unpack_build',
        'clusterfuzz._internal.fuzzing.fuzzer_selection.get_fuzz_target_weights',
        'clusterfuzz._internal.system.shell.clear_temp_directory',
        'time.time',
    ])

    test_helpers.patch_environ(self)

    os.environ['BUILDS_DIR'] = '/builds'
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['APP_NAME'] = FAKE_APP_NAME
    os.environ['JOB_NAME'] = 'job'

    self.mock._unpack_build.side_effect = _mock_unpack_build

  def _assert_env_vars(self):
    """Assert env vars exist."""
    self.assertEqual(os.environ['BUILD_URL'], 'gs://path/file-release-2.zip')

    self.assertEqual(
        os.environ['APP_PATH'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/revisions/app')

    self.assertEqual(
        os.environ['GN_ARGS_PATH'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/revisions/'
        'args.gn')

    self.assertEqual(
        os.environ['APP_DIR'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/revisions')

    self.assertEqual(
        os.environ['LLVM_SYMBOLIZER_PATH'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/revisions/'
        'llvm-symbolizer')

    self.assertEqual(
        os.environ['BUILD_DIR'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/revisions')

  def test_setup(self):
    """Tests setting up a build."""
    os.environ['RELEASE_BUILD_BUCKET_PATH'] = (
        'gs://path/file-release-([0-9]+).zip')

    self.mock.get_build_urls_list.return_value = [
        'gs://path/file-release-10.zip',
        'gs://path/file-release-2.zip',
        'gs://path/file-release-1.zip',
    ]

    self.mock.time.return_value = 1000.0
    build = build_manager.setup_regular_build(2)
    self.assertIsInstance(build, build_manager.RegularBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1000.0)

    self.mock._unpack_build.assert_called_once_with(
        mock.ANY, '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025',
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/revisions',
        'gs://path/file-release-2.zip', None)

    self._assert_env_vars()
    self.assertEqual(os.environ['APP_REVISION'], '2')

    self.mock.time.return_value = 1005.0
    self.assertIsInstance(
        build_manager.setup_regular_build(2), build_manager.RegularBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1005.0)

    # Already set up.
    self.assertEqual(self.mock._unpack_build.call_count, 1)
    self._assert_env_vars()
    self.assertEqual(os.environ['APP_REVISION'], '2')

    # Non-existent revisions do not result in any builds being set up.
    self.assertIsNone(build_manager.setup_regular_build(3))

  def test_delete(self):
    """Test deleting this build."""
    os.environ['RELEASE_BUILD_BUCKET_PATH'] = (
        'gs://path/file-release-([0-9]+).zip')

    self.mock.get_build_urls_list.return_value = [
        'gs://path/file-release-2.zip',
    ]

    build = build_manager.setup_regular_build(2)
    self.assertTrue(
        os.path.isdir(
            '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/revisions'))
    build.delete()
    self.assertFalse(
        os.path.isdir(
            '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/revisions'))
    self.assertTrue(
        os.path.isdir('/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025'))


class RegularLibFuzzerBuildTest(fake_filesystem_unittest.TestCase):
  """Tests for regular libFuzzer build setup."""

  def setUp(self):
    """Setup for regular libFuzzer build test."""
    test_utils.set_up_pyfakefs(self)

    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.utils.get_fuzz_targets',
        'clusterfuzz._internal.build_management.build_manager.get_build_urls_list',
        'clusterfuzz._internal.build_management.build_manager.Build.'
        '_get_fuzz_targets_from_archive',
        'clusterfuzz._internal.build_management.build_manager._make_space',
        'clusterfuzz._internal.build_management.build_manager._make_space_for_build',
        'clusterfuzz._internal.system.shell.clear_temp_directory',
        'clusterfuzz._internal.google_cloud_utils.storage.copy_file_from',
        'clusterfuzz._internal.google_cloud_utils.storage.get_download_file_size',
        'clusterfuzz._internal.system.archive.unpack',
        'time.time',
    ])

    test_helpers.patch_environ(self)

    os.environ['BUILDS_DIR'] = '/builds'
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['APP_NAME'] = FAKE_APP_NAME
    os.environ['JOB_NAME'] = 'libfuzzer_job'

    self.target_weights = {
        'target1': 0.0,
        'target2': 1.0,
        'target3': 0.0,
    }
    self.mock.get_download_file_size.return_value = 1
    self.mock.copy_file_from.return_value = True

    mock_targets_from_archive = self.mock._get_fuzz_targets_from_archive
    mock_targets_from_archive.return_value.__iter__.return_value = [
        'target1', 'target2', 'target3'
    ]
    self.mock.get_fuzz_targets.return_value = [
        '/path/target1', '/path/target2', '/path/target3'
    ]
    self.mock._make_space.return_value = True
    self.mock._make_space_for_build.return_value = True
    self.mock.unpack.return_value = True
    self.mock.time.return_value = 1000.0

    os.environ['RELEASE_BUILD_BUCKET_PATH'] = (
        'gs://path/file-release-([0-9]+).zip')

    self.mock.get_build_urls_list.return_value = [
        'gs://path/file-release-10.zip',
        'gs://path/file-release-2.zip',
        'gs://path/file-release-1.zip',
    ]

  def _assert_env_vars(self):
    self.assertEqual(os.environ['BUILD_URL'], 'gs://path/file-release-2.zip')
    self.assertEqual(
        os.environ['BUILD_DIR'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/revisions')

  @parameterized.parameterized.expand(['True', 'False'])
  def test_setup_fuzz(self, unpack_all):
    """Tests setting up a build during fuzzing."""
    os.environ['UNPACK_ALL_FUZZ_TARGETS_AND_FILES'] = unpack_all
    os.environ['TASK_NAME'] = 'fuzz'

    self.mock.time.return_value = 1000.0
    build = build_manager.setup_regular_build(
        2, target_weights=self.target_weights)
    self.assertIsInstance(build, build_manager.RegularBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1000.0)

    self.assertEqual('target2', os.environ['FUZZ_TARGET'])
    self._assert_env_vars()
    self.assertEqual(os.environ['APP_REVISION'], '2')

    self.assertEqual(1, self.mock.unpack.call_count)

    # Test setting up build again.
    os.environ['FUZZ_TARGET'] = ''
    self.mock.time.return_value = 1005.0
    self.assertIsInstance(
        build_manager.setup_regular_build(
            2, target_weights=self.target_weights), build_manager.RegularBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1005.0)

    # If it was a partial build, the unpack should be called again.
    if unpack_all == 'True':
      self.assertEqual(1, self.mock.unpack.call_count)
    else:
      self.assertEqual(2, self.mock.unpack.call_count)

    self.assertEqual('target2', os.environ['FUZZ_TARGET'])
    self.assertEqual('3', os.environ['FUZZ_TARGET_COUNT'])

  @parameterized.parameterized.expand(['True', 'False'])
  def test_setup_nonfuzz(self, unpack_all):
    """Test setting up a build during a non-fuzz task."""
    os.environ['UNPACK_ALL_FUZZ_TARGETS_AND_FILES'] = unpack_all
    os.environ['TASK_NAME'] = 'progression'
    os.environ['FUZZ_TARGET'] = 'target3'

    build = build_manager.setup_regular_build(2)
    self.assertIsInstance(build, build_manager.RegularBuild)

    self.assertEqual(1, self.mock.unpack.call_count)
    if unpack_all == 'True':
      self.assertIsNone(build_manager._get_file_match_callback())
    else:
      self.assertIsNotNone(build_manager._get_file_match_callback())

    class FileMatchCallbackChecker(object):
      """Used to verify that the callback passed to unpack is what we expect."""

      def __eq__(_, file_match_callback):  # pylint: disable=no-self-argument
        if unpack_all == 'True':
          # Ensure that |file_match_callback| is always None when we are
          # unpacking everything.
          self.assertIsNone(file_match_callback)
          return True

        self.assertIsNotNone(file_match_callback)
        self.assertEqual(file_match_callback.__name__, 'file_match_callback')
        self.assertTrue(isinstance(file_match_callback, types.FunctionType))
        return True

    file_match_callback_checker = FileMatchCallbackChecker()
    self.mock.unpack.assert_called_with(
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/'
        'revisions/file-release-2.zip',
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/revisions',
        file_match_callback=file_match_callback_checker,
        trusted=True)
    self.assertEqual('target3', os.environ['FUZZ_TARGET'])

    build = build_manager.setup_regular_build(2)
    self.assertIsInstance(build, build_manager.RegularBuild)

    # If it was a partial build, the unpack should be called again.
    if unpack_all != 'True':
      self.assertEqual(2, self.mock.unpack.call_count)
      self.mock.unpack.assert_called_with(
          '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/'
          'revisions/file-release-2.zip',
          '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/revisions',
          file_match_callback=file_match_callback_checker,
          trusted=True)

    self.assertEqual('target3', os.environ['FUZZ_TARGET'])

  def test_delete(self):
    """Test deleting this build."""
    os.environ['FUZZ_TARGET'] = 'fuzz_target'
    os.environ['RELEASE_BUILD_BUCKET_PATH'] = (
        'gs://path/file-release-([0-9]+).zip')

    self.mock.get_build_urls_list.return_value = [
        'gs://path/file-release-2.zip',
    ]

    build = build_manager.setup_regular_build(2)
    self.assertTrue(
        os.path.isdir(
            '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/revisions'))
    build.delete()
    self.assertFalse(
        os.path.isdir(
            '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/revisions'))
    self.assertTrue(
        os.path.isdir('/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025'))


class SymbolizedBuildTest(fake_filesystem_unittest.TestCase):
  """Tests for symbolized build setup."""

  def setUp(self):
    """Setup for symbolized build test."""
    test_utils.set_up_pyfakefs(self)

    test_helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager.get_build_urls_list',
        'clusterfuzz._internal.build_management.build_manager.Build._unpack_build',
        'clusterfuzz._internal.system.shell.clear_temp_directory', 'time.time'
    ])

    test_helpers.patch_environ(self)

    os.environ['BUILDS_DIR'] = '/builds'
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['APP_NAME'] = FAKE_APP_NAME
    os.environ['JOB_NAME'] = 'job'

    self.release_urls = None
    self.debug_urls = None

    self.mock._unpack_build.side_effect = _mock_unpack_build
    self.mock.get_build_urls_list.side_effect = self._mock_get_build_urls_list

  def _mock_get_build_urls_list(self, bucket_path):
    if 'release' in bucket_path:
      return self.release_urls or []

    return self.debug_urls or []

  def _prepare_test(self, release_urls, debug_urls):
    """Prepare test."""
    os.environ['SYM_RELEASE_BUILD_BUCKET_PATH'] = ''
    os.environ['SYM_DEBUG_BUILD_BUCKET_PATH'] = ''

    if release_urls:
      os.environ['SYM_RELEASE_BUILD_BUCKET_PATH'] = (
          'gs://path/file-release-([0-9]+).zip')

    if debug_urls:
      os.environ['SYM_DEBUG_BUILD_BUCKET_PATH'] = (
          'gs://path/file-debug-([0-9]+).zip')

    self.release_urls = release_urls
    self.debug_urls = debug_urls

  def _assert_env_vars_both(self):
    """Assert env vars."""
    self.assertEqual(os.environ['BUILD_URL'], 'gs://path/file-release-2.zip')

    self.assertEqual(
        os.environ['APP_PATH'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/symbolized/'
        'release/app')

    self.assertEqual(
        os.environ['APP_PATH_DEBUG'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/symbolized/'
        'debug/app')

    self.assertEqual(
        os.environ['APP_DIR'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/symbolized/'
        'debug')

    self.assertEqual(
        os.environ['LLVM_SYMBOLIZER_PATH'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/symbolized/'
        'debug/llvm-symbolizer')

    self.assertEqual(
        os.environ['BUILD_DIR'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/symbolized/'
        'debug')

  def _assert_env_vars_release(self):
    """Assert env vars for release."""
    self.assertEqual(os.environ['BUILD_URL'], 'gs://path/file-release-2.zip')

    self.assertEqual(
        os.environ['APP_PATH'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/symbolized/'
        'release/app')
    self.assertEqual(os.environ['APP_PATH_DEBUG'], '')

    self.assertEqual(
        os.environ['APP_DIR'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/symbolized/'
        'release')

    self.assertEqual(
        os.environ['LLVM_SYMBOLIZER_PATH'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/symbolized/'
        'release/llvm-symbolizer')

    self.assertEqual(
        os.environ['BUILD_DIR'],
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/symbolized/'
        'release')

  def test_setup_both(self):
    """Tests setting up both release and debug builds."""
    self._prepare_test([
        'gs://path/file-release-10.zip',
        'gs://path/file-release-2.zip',
        'gs://path/file-release-1.zip',
    ], [
        'gs://path/file-debug-10.zip',
        'gs://path/file-debug-2.zip',
        'gs://path/file-debug-1.zip',
    ])

    self.mock.time.return_value = 1000.0
    build = build_manager.setup_symbolized_builds(2)
    self.assertIsInstance(build, build_manager.SymbolizedBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1000.0)

    self.mock._unpack_build.assert_has_calls([
        mock.call(
            mock.ANY, '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025',
            '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/'
            'symbolized/release', 'gs://path/file-release-2.zip'),
        mock.call(
            mock.ANY, '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025',
            '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/'
            'symbolized/debug', 'gs://path/file-debug-2.zip'),
    ])
    self._assert_env_vars_both()
    self.assertEqual(os.environ['APP_REVISION'], '2')

    self.mock.time.return_value = 1005.0
    build = build_manager.setup_symbolized_builds(2)
    self.assertIsInstance(build, build_manager.SymbolizedBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1005.0)

    self._assert_env_vars_both()
    self.assertEqual(os.environ['APP_REVISION'], '2')
    self.assertTrue(self.mock._unpack_build.call_count, 2)

    self.assertIsNone(build_manager.setup_symbolized_builds(4))

  def test_setup_release_only(self):
    """Tests setting up release builds."""
    self._prepare_test([
        'gs://path/file-release-10.zip',
        'gs://path/file-release-2.zip',
        'gs://path/file-release-1.zip',
    ], None)

    self.mock.time.return_value = 1000.0
    build = build_manager.setup_symbolized_builds(2)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1000.0)

    self.assertIsInstance(build, build_manager.SymbolizedBuild)
    self.mock._unpack_build.assert_called_once_with(
        mock.ANY, '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025',
        '/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/'
        'symbolized/release', 'gs://path/file-release-2.zip')
    self._assert_env_vars_release()
    self.assertEqual(os.environ['APP_REVISION'], '2')

    self.mock.time.return_value = 1005.0
    self.assertIsInstance(
        build_manager.setup_symbolized_builds(2), build_manager.SymbolizedBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1005.0)
    self._assert_env_vars_release()
    self.assertEqual(os.environ['APP_REVISION'], '2')
    self.assertTrue(self.mock._unpack_build.call_count, 1)

    self.assertIsNone(build_manager.setup_symbolized_builds(4))

  def test_delete(self):
    """Test deleting this build."""
    self._prepare_test([
        'gs://path/file-release-10.zip',
        'gs://path/file-release-2.zip',
        'gs://path/file-release-1.zip',
    ], [
        'gs://path/file-debug-10.zip',
        'gs://path/file-debug-2.zip',
        'gs://path/file-debug-1.zip',
    ])

    build = build_manager.setup_symbolized_builds(2)
    self.assertTrue(
        os.path.isdir('/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/'
                      'symbolized/release'))
    self.assertTrue(
        os.path.isdir('/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/'
                      'symbolized/debug'))
    build.delete()
    self.assertFalse(
        os.path.isdir('/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/'
                      'symbolized/release'))
    self.assertFalse(
        os.path.isdir('/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025/'
                      'symbolized/debug'))
    self.assertTrue(
        os.path.isdir('/builds/path_be4c9ca0267afcd38b7c1a3eebb5998d0908f025'))


class ProductionBuildTest(fake_filesystem_unittest.TestCase):
  """Tests for production build setup."""

  def setUp(self):
    """Setup for production build test."""
    test_utils.set_up_pyfakefs(self)

    test_helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager.get_build_urls_list',
        'clusterfuzz._internal.build_management.build_manager.Build._unpack_build',
        'clusterfuzz._internal.system.shell.clear_temp_directory',
        'time.sleep',
        'time.time',
    ])

    test_helpers.patch_environ(self)

    os.environ['BUILDS_DIR'] = '/builds'
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['APP_NAME'] = FAKE_APP_NAME
    os.environ['JOB_NAME'] = 'job'
    os.environ['VERSION_PATTERN'] = r'.*-([0-9.]+).zip'

    os.environ['STABLE_BUILD_BUCKET_PATH'] = (
        'gs://path/file-stable-([0-9.]+).zip')
    os.environ['BETA_BUILD_BUCKET_PATH'] = ('gs://path/file-beta-([0-9.]+).zip')

    self.mock._unpack_build.side_effect = _mock_unpack_build
    self.mock.get_build_urls_list.side_effect = self._mock_get_build_urls_list

  def _mock_get_build_urls_list(self, bucket_path):
    """Mock get_build_urls_list()"""
    if not bucket_path:
      return []

    if 'extended_stable' in bucket_path:
      return [
          'gs://path/file-extended_stable-45.0.1824.2.zip',
          'gs://path/file-extended_stable-44.0.1824.1.zip',
          'gs://path/file-extended_stable-44.0.1822.2.zip',
      ]

    if 'stable' in bucket_path:
      return [
          'gs://path/file-stable-45.0.1824.2.zip',
          'gs://path/file-stable-44.0.1824.1.zip',
          'gs://path/file-stable-44.0.1822.2.zip',
      ]

    return [
        'gs://path/file-beta-45.0.1824.2.zip',
        'gs://path/file-beta-44.0.1824.1.zip',
        'gs://path/file-beta-44.0.1822.2.zip',
    ]

  def _assert_env_vars(self, build_type):
    """Assert env vars."""
    self.assertEqual(os.environ['BUILD_URL'],
                     'gs://path/file-%s-45.0.1824.2.zip' % build_type)

    self.assertEqual(
        os.environ['APP_PATH'],
        '/builds/path_8102046d3cea496c945743eb5f79284e7b10b51b/%s/app' %
        build_type)

    self.assertEqual(os.environ['APP_PATH_DEBUG'], '')

    self.assertEqual(
        os.environ['APP_DIR'],
        '/builds/path_8102046d3cea496c945743eb5f79284e7b10b51b/%s' % build_type)

    self.assertEqual(
        os.environ['LLVM_SYMBOLIZER_PATH'],
        '/builds/path_8102046d3cea496c945743eb5f79284e7b10b51b/%s/'
        'llvm-symbolizer' % build_type)

    self.assertEqual(
        os.environ['BUILD_DIR'],
        '/builds/path_8102046d3cea496c945743eb5f79284e7b10b51b/%s' % build_type)

  def test_setup_stable(self):
    """Test setting up a stable build."""
    self.mock.time.return_value = 1000.0
    build = build_manager.setup_production_build('stable')
    self.assertIsInstance(build, build_manager.ProductionBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1000.0)

    self.assertEqual(build.revision, '45.0.1824.2')
    self.assertEqual(os.environ['APP_REVISION'], '45.0.1824.2')
    self._assert_env_vars('stable')

    self.mock._unpack_build.assert_called_once_with(
        mock.ANY, '/builds/path_8102046d3cea496c945743eb5f79284e7b10b51b',
        '/builds/path_8102046d3cea496c945743eb5f79284e7b10b51b/stable',
        'gs://path/file-stable-45.0.1824.2.zip')

    self.mock.time.return_value = 1005.0
    self.assertEqual(
        build_manager.setup_production_build('stable').revision, '45.0.1824.2')
    self.assertEqual(_get_timestamp(build.base_build_dir), 1005.0)
    self._assert_env_vars('stable')
    self.assertEqual(self.mock._unpack_build.call_count, 1)

    self.assertIsNone(build_manager.setup_production_build('wrong'))

  def test_setup_beta(self):
    """Test setting up a stable build."""
    self.mock.time.return_value = 1000.0
    build = build_manager.setup_production_build('beta')
    self.assertIsInstance(build, build_manager.ProductionBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1000.0)
    self.assertEqual(os.environ['APP_REVISION'], '45.0.1824.2')
    self._assert_env_vars('beta')

    self.mock._unpack_build.assert_called_once_with(
        mock.ANY, '/builds/path_8102046d3cea496c945743eb5f79284e7b10b51b',
        '/builds/path_8102046d3cea496c945743eb5f79284e7b10b51b/beta',
        'gs://path/file-beta-45.0.1824.2.zip')

    self.mock.time.return_value = 1005.0
    self.assertEqual(
        build_manager.setup_production_build('beta').revision, '45.0.1824.2')
    self.assertEqual(_get_timestamp(build.base_build_dir), 1005.0)
    self._assert_env_vars('beta')
    self.assertEqual(self.mock._unpack_build.call_count, 1)

  def test_setup_extended_stable(self):
    """Test setting up an extended stable build."""
    os.environ['EXTENDED_STABLE_BUILD_BUCKET_PATH'] = (
        'gs://path/file-extended_stable-([0-9.]+).zip')

    self.mock.time.return_value = 1000.0
    build = build_manager.setup_production_build('extended_stable')
    self.assertIsInstance(build, build_manager.ProductionBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1000.0)

    self.assertEqual(build.revision, '45.0.1824.2')
    self.assertEqual(os.environ['APP_REVISION'], '45.0.1824.2')
    self._assert_env_vars('extended_stable')

    self.mock._unpack_build.assert_called_once_with(
        mock.ANY, '/builds/path_8102046d3cea496c945743eb5f79284e7b10b51b',
        '/builds/path_8102046d3cea496c945743eb5f79284e7b10b51b/extended_stable',
        'gs://path/file-extended_stable-45.0.1824.2.zip')

    self.mock.time.return_value = 1005.0
    self.assertEqual(
        build_manager.setup_production_build('extended_stable').revision,
        '45.0.1824.2')
    self.assertEqual(_get_timestamp(build.base_build_dir), 1005.0)
    self._assert_env_vars('extended_stable')
    self.assertEqual(self.mock._unpack_build.call_count, 1)

    self.assertIsNone(build_manager.setup_production_build('wrong'))

  def test_delete(self):
    """Test deleting this build."""
    build = build_manager.setup_production_build('stable')
    self.assertTrue(
        os.path.isdir(
            '/builds/path_8102046d3cea496c945743eb5f79284e7b10b51b/stable'))
    build.delete()
    self.assertFalse(
        os.path.isdir(
            '/builds/path_8102046d3cea496c945743eb5f79284e7b10b51b/stable'))
    self.assertTrue(
        os.path.isdir('/builds/path_8102046d3cea496c945743eb5f79284e7b10b51b'))


@test_utils.with_cloud_emulators('datastore')
class CustomBuildTest(fake_filesystem_unittest.TestCase):
  """Tests for custom build setup."""

  def setUp(self):
    """Setup for custom build test."""
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager._make_space_for_build',
        'clusterfuzz._internal.system.shell.clear_temp_directory',
        'clusterfuzz._internal.google_cloud_utils.blobs.read_blob_to_disk',
        'clusterfuzz._internal.system.archive.unpack',
        'time.sleep',
        'time.time',
    ])

    os.environ['BUILDS_DIR'] = '/builds'
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['APP_NAME'] = FAKE_APP_NAME
    os.environ['CUSTOM_BINARY'] = 'True'

    data_types.Job(
        name='job_custom',
        custom_binary_key='key',
        custom_binary_filename='custom_binary.zip',
        custom_binary_revision=3).put()

    test_utils.set_up_pyfakefs(self)
    self.mock._make_space_for_build.return_value = True
    self.mock.unpack.side_effect = self._mock_unpack
    self.mock.read_blob_to_disk.return_value = True

  # pylint: disable=unused-argument
  def _mock_unpack(self, _, build_dir, trusted=True):
    """mock archive.unpack."""
    _mock_unpack_build(None, None, build_dir, None)

  def _assert_env_vars(self):
    """Assert env vars."""
    self.assertEqual(os.environ['APP_PATH'], '/builds/job_custom/custom/app')
    self.assertEqual(os.environ['APP_DIR'], '/builds/job_custom/custom')
    self.assertEqual(os.environ['APP_PATH_DEBUG'], '')

    self.assertEqual(os.environ['LLVM_SYMBOLIZER_PATH'],
                     '/builds/job_custom/custom/llvm-symbolizer')

    self.assertEqual(os.environ['APP_REVISION'], '3')
    self.assertEqual(os.environ['BUILD_KEY'], 'key')
    self.assertEqual(os.environ['BUILD_DIR'], '/builds/job_custom/custom')

  def test_setup(self):
    """Test setting up a custom binary."""
    os.environ['JOB_NAME'] = 'job_custom'
    self.mock.time.return_value = 1000.0
    build = build_manager.setup_custom_binary()
    self.assertIsInstance(build, build_manager.CustomBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1000.0)

    self.mock.read_blob_to_disk.assert_called_once_with(
        'key', '/builds/job_custom/custom/custom_binary.zip')

    self.mock.unpack.assert_called_once_with(
        '/builds/job_custom/custom/custom_binary.zip',
        '/builds/job_custom/custom',
        trusted=True)

    self._assert_env_vars()

    self.mock.time.return_value = 1005.0
    self.assertIsInstance(build_manager.setup_custom_binary(),
                          build_manager.CustomBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1005.0)
    self.assertEqual(self.mock.read_blob_to_disk.call_count, 1)
    self.assertEqual(self.mock.unpack.call_count, 1)
    self._assert_env_vars()

  def test_setup_shared(self):
    """Test setting up a custom binary (shared)."""
    os.environ['JOB_NAME'] = 'job_share'
    os.environ['SHARE_BUILD_WITH_JOB_TYPE'] = 'job_custom'

    self.mock.time.return_value = 1000.0
    build = build_manager.setup_custom_binary()
    self.assertIsInstance(build, build_manager.CustomBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1000.0)

    self.mock.read_blob_to_disk.assert_called_once_with(
        'key', '/builds/job_custom/custom/custom_binary.zip')

    self.mock.unpack.assert_called_once_with(
        '/builds/job_custom/custom/custom_binary.zip',
        '/builds/job_custom/custom',
        trusted=True)

    self._assert_env_vars()
    self.assertEqual(os.environ['JOB_NAME'], 'job_share')

  def test_delete(self):
    """Test deleting this build."""
    os.environ['JOB_NAME'] = 'job_custom'
    build = build_manager.setup_custom_binary()
    self.assertTrue(os.path.isdir('/builds/job_custom/custom'))
    build.delete()
    self.assertFalse(os.path.isdir('/builds/job_custom/custom'))
    self.assertTrue(os.path.isdir('/builds/job_custom'))


class SystemBuildTest(fake_filesystem_unittest.TestCase):
  """Tests for system build setup."""

  def setUp(self):
    """Setup for system build test."""
    test_utils.set_up_pyfakefs(self)

    test_helpers.patch(self, [
        'clusterfuzz._internal.system.shell.clear_temp_directory',
    ])

    test_helpers.patch_environ(self)

    os.environ['BUILDS_DIR'] = '/builds'
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['APP_NAME'] = FAKE_APP_NAME
    os.environ['SYSTEM_BINARY_DIR'] = '/system_binary'
    _mock_unpack_build(
        build_manager.SystemBuild('/'), None, '/system_binary', None)

  def test_setup(self):
    """Test setting up a system binary."""
    self.assertIsNotNone(build_manager.setup_system_binary())

    self.assertEqual(os.environ['APP_REVISION'], '1')
    self.assertEqual(os.environ['APP_PATH'], '/system_binary/app')
    self.assertEqual(os.environ['APP_DIR'], '/system_binary')
    self.assertEqual(os.environ['BUILD_DIR'], '/system_binary')

  def test_delete(self):
    """Test deleting this build."""
    build = build_manager.setup_system_binary()
    with self.assertRaises(build_manager.BuildManagerException):
      build.delete()


class AuxiliaryRegularBuildTest(fake_filesystem_unittest.TestCase):
  """Tests for auxiliary build setup."""

  def setUp(self):
    """Setup for auxiliary regular build Test."""
    test_utils.set_up_pyfakefs(self)

    test_helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager.get_build_urls_list',
        'clusterfuzz._internal.build_management.build_manager.Build._unpack_build',
        'clusterfuzz._internal.system.shell.clear_temp_directory',
        'time.time',
    ])

    test_helpers.patch_environ(self)

    os.environ['BUILDS_DIR'] = '/builds'
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['APP_NAME'] = FAKE_APP_NAME
    os.environ['JOB_NAME'] = 'job'

    self.mock._unpack_build.side_effect = _mock_unpack_build

  def _assert_env_vars(self):
    """An auxiliary build show not overwrite the common env variables, but
    should set up some variables with a prefix in their names."""
    self.assertIsNone(os.environ.get('APP_PATH'))
    self.assertIsNone(os.environ.get('APP_REVISION'))
    self.assertIsNone(os.environ.get('BUILD_DIR'))
    self.assertIsNone(os.environ.get('BUILD_URL'))
    self.assertIsNone(os.environ.get('GN_ARGS_PATH'))

    self.assertEqual(os.environ['DATAFLOW_APP_PATH'], '')
    self.assertEqual(
        os.environ['DATAFLOW_BUILD_DIR'],
        '/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d/dataflow')
    self.assertEqual(os.environ['DATAFLOW_BUILD_URL'],
                     'gs://path/file-dataflow-10.zip')

  def test_setup(self):
    """Tests setting up a build."""
    os.environ['DATAFLOW_BUILD_BUCKET_PATH'] = (
        'gs://path/file-dataflow-([0-9]+).zip')

    self.mock.get_build_urls_list.return_value = [
        'gs://path/file-dataflow-10.zip',
        'gs://path/file-dataflow-2.zip',
        'gs://path/file-dataflow-1.zip',
    ]

    self.mock.time.return_value = 1000.0
    build = build_manager.setup_trunk_build(
        [os.environ['DATAFLOW_BUILD_BUCKET_PATH']], build_prefix='DATAFLOW')
    self.assertIsInstance(build, build_manager.RegularBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1000.0)

    self.mock._unpack_build.assert_called_once_with(
        mock.ANY, '/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d',
        '/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d/dataflow',
        'gs://path/file-dataflow-10.zip', None)

    self._assert_env_vars()

    self.mock.time.return_value = 1005.0
    self.assertIsInstance(
        build_manager.setup_trunk_build(
            [os.environ['DATAFLOW_BUILD_BUCKET_PATH']],
            build_prefix='DATAFLOW'), build_manager.RegularBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1005.0)

    # Already set up.
    self.assertEqual(self.mock._unpack_build.call_count, 1)
    self._assert_env_vars()

  def test_delete(self):
    """Test deleting this build."""
    os.environ['DATAFLOW_BUILD_BUCKET_PATH'] = (
        'gs://path/file-dataflow-([0-9]+).zip')

    self.mock.get_build_urls_list.return_value = [
        'gs://path/file-dataflow-2.zip',
    ]

    build = fuzz_task.build_manager.setup_trunk_build(
        [os.environ['DATAFLOW_BUILD_BUCKET_PATH']], 'DATAFLOW')
    self.assertTrue(
        os.path.isdir(
            '/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d/revisions'))
    build.delete()
    self.assertFalse(
        os.path.isdir(
            '/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d/dataflow'))
    self.assertTrue(
        os.path.isdir('/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d'))


class AuxiliaryRegularLibFuzzerBuildTest(fake_filesystem_unittest.TestCase):
  """Tests for auxiliary libFuzzer build setup."""

  def setUp(self):
    """Setup for auxiliary regular libFuzzer build test."""
    test_utils.set_up_pyfakefs(self)

    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.utils.get_fuzz_targets',
        'clusterfuzz._internal.build_management.build_manager.get_build_urls_list',
        'clusterfuzz._internal.build_management.build_manager.Build.'
        '_get_fuzz_targets_from_archive',
        'clusterfuzz._internal.build_management.build_manager._make_space',
        'clusterfuzz._internal.build_management.build_manager._make_space_for_build',
        'clusterfuzz._internal.system.shell.clear_temp_directory',
        'clusterfuzz._internal.google_cloud_utils.storage.copy_file_from',
        'clusterfuzz._internal.google_cloud_utils.storage.get_download_file_size',
        'clusterfuzz._internal.system.archive.unpack',
        'time.time',
    ])

    test_helpers.patch_environ(self)

    os.environ['BUILDS_DIR'] = '/builds'
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['APP_NAME'] = FAKE_APP_NAME
    os.environ['JOB_NAME'] = 'libfuzzer_job'

    self.target_weights = {
        'target1': 0.0,
        'target2': 1.0,
        'target3': 0.0,
    }
    self.mock.get_download_file_size.return_value = 1
    self.mock.copy_file_from.return_value = True

    mock_targets_from_archive = self.mock._get_fuzz_targets_from_archive
    mock_targets_from_archive.return_value.__iter__.return_value = [
        'target1', 'target2', 'target3'
    ]
    self.mock.get_fuzz_targets.return_value = [
        '/path/target1', '/path/target2', '/path/target3'
    ]
    self.mock._make_space.return_value = True
    self.mock._make_space_for_build.return_value = True
    self.mock.unpack.return_value = True
    self.mock.time.return_value = 1000.0

    os.environ['DATAFLOW_BUILD_BUCKET_PATH'] = (
        'gs://path/file-dataflow-([0-9]+).zip')

    self.mock.get_build_urls_list.return_value = [
        'gs://path/file-dataflow-10.zip',
        'gs://path/file-dataflow-2.zip',
        'gs://path/file-dataflow-1.zip',
    ]

  def _assert_env_vars(self):
    # Assert the expected values of environment variables.
    self.assertIsNone(os.environ.get('APP_REVISION'))
    self.assertIsNone(os.environ.get('BUILD_URL'))

    self.assertEqual(os.environ['DATAFLOW_APP_REVISION'], '10')
    self.assertEqual(
        os.environ['DATAFLOW_BUILD_DIR'],
        '/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d/dataflow')
    self.assertEqual(os.environ['DATAFLOW_BUILD_URL'],
                     'gs://path/file-dataflow-10.zip')

  @parameterized.parameterized.expand(['True', 'False'])
  def test_setup_fuzz(self, unpack_all):
    """Tests setting up a build during fuzzing."""
    os.environ['UNPACK_ALL_FUZZ_TARGETS_AND_FILES'] = unpack_all
    os.environ['TASK_NAME'] = 'fuzz'
    self.mock.time.return_value = 1000.0

    class FileMatchCallbackChecker(object):
      """Used to verify that the callback passed to unpack is what we expect."""

      def __eq__(_, file_match_callback):  # pylint: disable=no-self-argument
        if unpack_all == 'True':
          # Ensure that |file_match_callback| is always None when we are
          # unpacking everything.
          self.assertIsNone(file_match_callback)
          return True

        self.assertIsNotNone(file_match_callback)
        self.assertEqual(file_match_callback.__name__, 'file_match_callback')
        self.assertTrue(isinstance(file_match_callback, types.FunctionType))
        return True

    # Test setting up build with a fuzz target specified (comes from regular
    # build setup).
    os.environ['FUZZ_TARGET'] = 'target1'

    if unpack_all == 'True':
      self.assertIsNone(build_manager._get_file_match_callback())
    else:
      self.assertIsNotNone(build_manager._get_file_match_callback())

    build = build_manager.setup_trunk_build(
        [os.environ['DATAFLOW_BUILD_BUCKET_PATH']],
        build_prefix='DATAFLOW',
        target_weights=self.target_weights)
    self.assertIsInstance(build, build_manager.RegularBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1000.0)
    self.assertEqual('target1', os.environ['FUZZ_TARGET'])
    self._assert_env_vars()

    self.assertEqual(1, self.mock.unpack.call_count)
    file_match_callback_checker = FileMatchCallbackChecker()
    self.mock.unpack.assert_called_with(
        '/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d/'
        'dataflow/file-dataflow-10.zip',
        '/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d/dataflow',
        file_match_callback=file_match_callback_checker,
        trusted=True)

    os.environ['FUZZ_TARGET'] = 'target3'
    self.mock.time.return_value = 1005.0
    self.assertIsInstance(
        build_manager.setup_trunk_build(
            [os.environ['DATAFLOW_BUILD_BUCKET_PATH']],
            build_prefix='DATAFLOW',
            target_weights=self.target_weights), build_manager.RegularBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1005.0)

    # An auxiliary build set up should not choose and overwrite the fuzz target.
    self.assertEqual('target3', os.environ['FUZZ_TARGET'])

    # If it was a partial build, the unpack should be called again.
    if unpack_all == 'True':
      self.assertEqual(1, self.mock.unpack.call_count)
    else:
      self.assertEqual(2, self.mock.unpack.call_count)

    self.mock.unpack.assert_called_with(
        '/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d/'
        'dataflow/file-dataflow-10.zip',
        '/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d/dataflow',
        file_match_callback=file_match_callback_checker,
        trusted=True)

  def test_delete(self):
    """Test deleting this build."""
    os.environ['FUZZ_TARGET'] = 'fuzz_target'
    os.environ['DATAFLOW_BUILD_BUCKET_PATH'] = (
        'gs://path/file-dataflow-([0-9]+).zip')

    self.mock.get_build_urls_list.return_value = [
        'gs://path/file-dataflow-2.zip',
    ]

    build = build_manager.setup_trunk_build(
        [os.environ['DATAFLOW_BUILD_BUCKET_PATH']], 'DATAFLOW')
    self.assertTrue(
        os.path.isdir(
            '/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d/dataflow'))
    build.delete()
    self.assertFalse(
        os.path.isdir(
            '/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d/dataflow'))
    self.assertTrue(
        os.path.isdir('/builds/path_2992e823e35fd34a63e0f8733cdafd6875036a1d'))


@mock.patch(
    'clusterfuzz._internal.build_management.build_manager.MAX_EVICTED_BUILDS',
    3)
class BuildEvictionTests(fake_filesystem_unittest.TestCase):
  """Build eviction tests."""

  def setUp(self):
    """Setup for build eviction tests."""
    test_utils.set_up_pyfakefs(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.base.utils.is_chromium',
        'clusterfuzz._internal.system.shell.get_free_disk_space',
        'clusterfuzz._internal.system.archive.extracted_size',
    ])

    test_helpers.patch_environ(self)

    os.environ['BUILDS_DIR'] = '/builds'
    os.environ['FAIL_RETRIES'] = '1'

    os.makedirs('/builds/build1/revisions')
    os.makedirs('/builds/build2/revisions')
    os.makedirs('/builds/build3/revisions')
    os.makedirs('/builds/build4/revisions')
    self.fs.create_file(
        '/builds/build1/.timestamp', contents='1486166114.668105')
    self.fs.create_file(
        '/builds/build2/.timestamp', contents='1486166110.142942')
    self.fs.create_file(
        '/builds/build3/.timestamp', contents='1486166112.180345')

    self.free_disk_space = []
    self.mock.is_chromium.return_value = True

  def _mock_free_disk_space(self, _):
    return self.free_disk_space.pop(0)

  def test_make_space_for_build_remove_one_build(self):
    """Test make_space_for_build (remove 1 build)."""
    self.mock.extracted_size.return_value = 1 * 1024 * 1024 * 1024  # 1 GB
    self.mock.get_free_disk_space.side_effect = self._mock_free_disk_space
    self.free_disk_space = [
        9 * 1024 * 1024 * 1024,
        24 * 1024 * 1024 * 1024,
    ]

    self.assertTrue(
        build_manager._make_space_for_build('/archive.zip', '/builds/build4'))

    self.assertTrue(os.path.isdir('/builds/build1'))
    self.assertFalse(os.path.isdir('/builds/build2'))
    self.assertTrue(os.path.isdir('/builds/build3'))
    self.assertTrue(os.path.isdir('/builds/build4'))

  def test_make_space_for_build_remove_two_builds(self):
    """Test make_space_for_build (remove 2 builds)."""
    self.mock.extracted_size.return_value = 1 * 1024 * 1024 * 1024  # 1 GB
    self.mock.get_free_disk_space.side_effect = self._mock_free_disk_space
    self.free_disk_space = [
        8 * 1024 * 1024 * 1024,
        9 * 1024 * 1024 * 1024,
        12 * 1024 * 1024 * 1024,
    ]

    self.assertTrue(
        build_manager._make_space_for_build('/archive.zip', '/builds/build4'))

    self.assertTrue(os.path.isdir('/builds/build1'))
    self.assertFalse(os.path.isdir('/builds/build2'))
    self.assertFalse(os.path.isdir('/builds/build3'))
    self.assertTrue(os.path.isdir('/builds/build4'))

  def test_make_space_for_build_remove_three_builds(self):
    """Test make_space_for_build (remove 3 builds)."""
    self.mock.extracted_size.return_value = 1 * 1024 * 1024 * 1024  # 1 GB
    self.mock.get_free_disk_space.side_effect = self._mock_free_disk_space
    self.free_disk_space = [
        7 * 1024 * 1024 * 1024,
        8 * 1024 * 1024 * 1024,
        9 * 1024 * 1024 * 1024,
        14 * 1024 * 1024 * 1024,
    ]

    self.assertTrue(
        build_manager._make_space_for_build('/archive.zip', '/builds/build4'))

    self.assertFalse(os.path.isdir('/builds/build1'))
    self.assertFalse(os.path.isdir('/builds/build2'))
    self.assertFalse(os.path.isdir('/builds/build3'))
    self.assertTrue(os.path.isdir('/builds/build4'))

  def test_make_space_for_build_fail(self):
    """Test make_space_for_build failure."""
    self.mock.extracted_size.return_value = 20 * 1024 * 1024 * 1024  # 1 GB
    self.mock.get_free_disk_space.side_effect = self._mock_free_disk_space
    self.free_disk_space = [
        12 * 1024 * 1024 * 1024,
        17 * 1024 * 1024 * 1024,
        18 * 1024 * 1024 * 1024,
        24 * 1024 * 1024 * 1024,
    ]

    self.assertFalse(
        build_manager._make_space_for_build('/archive.zip', '/builds/build4'))

    self.assertFalse(os.path.isdir('/builds/build1'))
    self.assertFalse(os.path.isdir('/builds/build2'))
    self.assertFalse(os.path.isdir('/builds/build3'))
    self.assertTrue(os.path.isdir('/builds/build4'))

  def test_make_space_for_build_no_builds_to_remove(self):
    """Test _make_space_for_build failure (no builds to remove)."""
    shutil.rmtree('/builds/build1')
    shutil.rmtree('/builds/build2')
    shutil.rmtree('/builds/build3')

    self.mock.extracted_size.return_value = 20 * 1024 * 1024 * 1024  # 1 GB
    self.mock.get_free_disk_space.side_effect = self._mock_free_disk_space
    self.free_disk_space = [
        18 * 1024 * 1024 * 1024,
    ]

    self.assertFalse(
        build_manager._make_space_for_build('/archive.zip', '/builds/build4'))


class GetFileMatchCallbackTest(unittest.TestCase):
  """Tests for _get_file_match_callback."""

  def setUp(self):
    test_helpers.patch_environ(self)

  def _test_helper(self, fuzz_target):
    """Test helper."""
    os.environ['FUZZ_TARGET'] = fuzz_target
    match_callback = build_manager._get_file_match_callback()
    self.assertIsNotNone(match_callback)

    # TODO(metzman): This test (and many others), use hardcoded paths, meaning
    # that the tests won't work on Windows. Convert them to using os.path.join.
    # Tests for absolute path.
    self.assertTrue(match_callback('/b/build/my_fuzzer'))
    self.assertTrue(match_callback('/b/build/my_fuzzer.dict'))
    self.assertTrue(match_callback('/b/build/my_fuzzer.options'))
    self.assertTrue(match_callback('/b/build/llvm-symbolizer'))
    self.assertTrue(match_callback('/b/build/icudtl.dat'))
    self.assertTrue(match_callback('/b/build/swiftshader/libGLESv2.so'))
    self.assertTrue(
        match_callback(
            '/b/build/instrumented_libraries/msan/lib/libgcrypt.so.11.8.2'))
    self.assertTrue(match_callback('/b/build/afl-fuzz'))
    # Windows
    self.assertTrue(match_callback('my_fuzzer.exe'))
    # Python
    self.assertTrue(match_callback('my_fuzzer.par'))

    # Tests for relative path.
    self.assertTrue(match_callback('build/my_fuzzer'))
    self.assertTrue(match_callback('build/my_fuzzer.dict'))
    self.assertTrue(match_callback('build/my_fuzzer.options'))
    self.assertTrue(match_callback('build/llvm-symbolizer'))
    self.assertTrue(match_callback('build/icudtl.dat'))
    self.assertTrue(match_callback('build/swiftshader/libGLESv2.so'))
    self.assertTrue(
        match_callback(
            'build/instrumented_libraries/msan/lib/libgcrypt.so.11.8.2'))
    self.assertTrue(match_callback(r'build/afl-fuzz'))

    # Tests for absolute path on Mac platform.
    self.assertTrue(
        match_callback(
            r'/b/build/my_fuzzer.dSYM/Contents/Resources/DWARF/some_dependency')
    )

    # Tests for relative path on Mac platform.
    self.assertTrue(
        match_callback(
            r'build/my_fuzzer.dSYM/Contents/Resources/DWARF/some_dependency'))

    # Tests for other fuzz target files with absolute paths.
    self.assertFalse(match_callback('/b/build/other_fuzzer'))
    self.assertFalse(match_callback('/b/build/other_fuzzer.options'))
    self.assertFalse(match_callback('/b/build/other_fuzzer.zip'))
    self.assertFalse(match_callback('/b/afl-build/other_fuzzer'))
    self.assertTrue(match_callback('/b/build/other_fuzzer.dict'))

    # Tests for other fuzz target files with relative paths.
    self.assertFalse(match_callback('build/other_fuzzer'))
    self.assertFalse(match_callback('build/other_fuzzer.options'))
    self.assertFalse(match_callback('build/other_fuzzer.zip'))
    self.assertFalse(match_callback('afl-build/other_fuzzer'))
    self.assertTrue(match_callback('build/other_fuzzer.dict'))

    # Tests for other fuzz target files with absolute path on Mac platform.
    self.assertFalse(
        match_callback(r'/b/build/other_fuzzer.dSYM/Contents/Resources/DWARF/'
                       'some_dependency'))

    # Tests for other fuzz target files with relative path on Mac platform.
    self.assertFalse(
        match_callback(
            r'build/other_fuzzer.dSYM/Contents/Resources/DWARF/some_dependency')
    )

    # Test for other fuzz target files with relative path to current directory.
    self.assertFalse(match_callback('./other_fuzzer'))

    # Test for a non-dependency file format.
    self.assertFalse(match_callback('README.txt'))

    # Tests for files that appear in Windows libFuzzer builds.
    self.assertTrue(match_callback(fuzz_target + '.exe'))
    self.assertTrue(match_callback(fuzz_target + '.exe.pdb'))
    self.assertTrue(match_callback('shared.dll'))
    self.assertTrue(match_callback('shared.dll.pdb'))
    self.assertFalse(match_callback('other_fuzzer.exe'))
    self.assertFalse(match_callback('other_fuzzer.exe.pdb'))
    # Test that other .par targets are not included.
    self.assertFalse(match_callback('other_fuzzer.par'))
    # Make sure we aren't excluding files that contain but don't end in
    # ".exe.pdb".
    self.assertTrue(match_callback('other_fuzzer.exe.pdbb'))

  def test_with_fuzz_target(self):
    self._test_helper('my_fuzzer')

  def test_without_fuzz_target(self):
    match_callback = build_manager._get_file_match_callback()
    self.assertIsNone(match_callback)


@test_utils.integration
class RpathsTest(unittest.TestCase):
  """Rpath patching tests."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager.Build._unpack_build',
        'clusterfuzz._internal.system.shell.clear_temp_directory',
    ])

    os.environ['JOB_NAME'] = 'linux_msan_test'
    os.environ['INSTRUMENTED_LIBRARIES_PATHS_MSAN_CHAINED'] = (
        '/msan/lib:/msan/usr/lib')
    os.environ['INSTRUMENTED_LIBRARIES_PATHS_MSAN_NO_ORIGINS'] = (
        '/msan-no-origins/lib:/msan-no-origins/usr/lib')
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['APP_NAME'] = 'app'

    self.base_build_dir = tempfile.mkdtemp()

  def tearDown(self):
    shutil.rmtree(self.base_build_dir, ignore_errors=True)

  # pylint: disable=unused-argument
  def mock_unpack_build(self, test_build_dir, actual_self, base_build_dir,
                        build_dir, url, target_weights):
    test_data_dir = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'build_manager_data',
        test_build_dir)

    shell.remove_directory(build_dir, recreate=False)
    shutil.copytree(test_data_dir, build_dir)
    return True

  def test_patch_rpaths_no_origins(self):
    """Tests that no-origins libraries are used."""
    self.mock._unpack_build.side_effect = functools.partial(
        self.mock_unpack_build, 'rpath_new')
    build = build_manager.RegularBuild(self.base_build_dir, 1337, 'no-origins')
    self.assertTrue(build.setup())

    self.assertEqual(
        os.path.join(self.base_build_dir, 'revisions', 'app'),
        os.environ['APP_PATH'])

    rpaths = build_manager.get_rpaths(os.environ['APP_PATH'])
    self.assertListEqual(['/msan-no-origins/lib', '/msan-no-origins/usr/lib'],
                         rpaths)

  def test_patch_rpaths_not_available(self):
    """Tests that rpaths aren't added when libs aren't available.."""
    os.environ['JOB_NAME'] = 'linux_asan_test'

    self.mock._unpack_build.side_effect = functools.partial(
        self.mock_unpack_build, 'rpath_new')
    build = build_manager.RegularBuild(self.base_build_dir, 1337,
                                       'chained-origins')
    self.assertTrue(build.setup())

    self.assertEqual(
        os.path.join(self.base_build_dir, 'revisions', 'app'),
        os.environ['APP_PATH'])

    rpaths = build_manager.get_rpaths(os.environ['APP_PATH'])
    self.assertListEqual([], rpaths)

  def test_patch_rpaths_prepend(self):
    """Tests patching rpaths to a binary that already has an rpath."""
    self.mock._unpack_build.side_effect = functools.partial(
        self.mock_unpack_build, 'rpath_prepend_to_existing')
    build = build_manager.RegularBuild(self.base_build_dir, 1337,
                                       'chained-origins')
    self.assertTrue(build.setup())

    self.assertEqual(
        os.path.join(self.base_build_dir, 'revisions', 'app'),
        os.environ['APP_PATH'])

    rpaths = build_manager.get_rpaths(os.environ['APP_PATH'])
    self.assertListEqual(['/msan/lib', '/msan/usr/lib', '$ORIGIN/.'], rpaths)

  def test_patch_rpaths_chrpath(self):
    """Tests patching rpaths to a binary using chrpath."""
    limit = build_manager.PATCHELF_SIZE_LIMIT
    build_manager.PATCHELF_SIZE_LIMIT = 0

    def cleanup():
      build_manager.PATCHELF_SIZE_LIMIT = limit

    self.addCleanup(cleanup)

    self.mock._unpack_build.side_effect = functools.partial(
        self.mock_unpack_build, 'rpath_prepend_to_existing')
    build = build_manager.RegularBuild(self.base_build_dir, 1337,
                                       'chained-origins')
    self.assertTrue(build.setup())

    self.assertEqual(
        os.path.join(self.base_build_dir, 'revisions', 'app'),
        os.environ['APP_PATH'])

    rpaths = build_manager.get_rpaths(os.environ['APP_PATH'])
    self.assertListEqual(['/msan/lib', '/msan/usr/lib', '$ORIGIN/.'], rpaths)

  def test_patch_rpaths_new(self):
    """Tests patching rpaths for a binary that doesn't have an rpath."""
    self.mock._unpack_build.side_effect = functools.partial(
        self.mock_unpack_build, 'rpath_new')
    build = build_manager.RegularBuild(self.base_build_dir, 1337,
                                       'chained-origins')
    self.assertTrue(build.setup())

    self.assertEqual(
        os.path.join(self.base_build_dir, 'revisions', 'app'),
        os.environ['APP_PATH'])

    rpaths = build_manager.get_rpaths(os.environ['APP_PATH'])
    self.assertListEqual(['/msan/lib', '/msan/usr/lib'], rpaths)

  def test_patch_rpaths_libfuzzer(self):
    """Tests patching rpaths for libFuzzer targets."""
    os.environ['JOB_NAME'] = 'libfuzzer_msan_test'

    self.mock._unpack_build.side_effect = functools.partial(
        self.mock_unpack_build, 'rpath_libfuzzer')
    build = build_manager.RegularBuild(self.base_build_dir, 1337,
                                       'chained-origins')
    self.assertTrue(build.setup())

    self.assertEqual('', os.environ['APP_DIR'])

    rpaths = build_manager.get_rpaths(
        os.path.join(os.environ['BUILD_DIR'], 'target_1'))
    self.assertListEqual(['/msan/lib', '/msan/usr/lib'], rpaths)

    rpaths = build_manager.get_rpaths(
        os.path.join(os.environ['BUILD_DIR'], 'target_2'))
    self.assertListEqual(['/msan/lib', '/msan/usr/lib'], rpaths)

  def test_patch_rpaths_existing_msan(self):
    """Tests patching rpaths for a binary that already has a msan rpath
    patched."""
    self.mock._unpack_build.side_effect = functools.partial(
        self.mock_unpack_build, 'rpath_existing_msan')
    build = build_manager.RegularBuild(self.base_build_dir, 1337,
                                       'chained-origins')
    self.assertTrue(build.setup())

    self.assertEqual(
        os.path.join(self.base_build_dir, 'revisions', 'app'),
        os.environ['APP_PATH'])

    rpaths = build_manager.get_rpaths(os.environ['APP_PATH'])
    self.assertListEqual(['/msan/lib', '/msan/usr/lib'], rpaths)


class SortBuildUrlsByRevisionTest(unittest.TestCase):
  """Test _sort_build_urls_by_revision."""

  def test_simple(self):
    """Tests regular case with and without reverse flag set."""
    bucket_path = ('gs://chromium-browser-libfuzzer/'
                   'linux-release-asan/libfuzzer-linux-release-([0-9]+).zip')
    build_urls = [
        'linux-release-asan/libfuzzer-linux-release-359936.zip',
        'linux-release-asan/libfuzzer-linux-release-359950.zip',
        'linux-release-asan/libfuzzer-linux-release-359945.zip',
        'linux-release-asan/libfuzzer-linux-release-359953.zip',
    ]
    expected_result = [
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-359953.zip',
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-359950.zip',
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-359945.zip',
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-359936.zip'
    ]
    actual_result = build_manager._sort_build_urls_by_revision(
        build_urls, bucket_path, reverse=True)
    self.assertEqual(expected_result, actual_result)

    expected_result = [
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-359936.zip',
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-359945.zip',
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-359950.zip',
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-359953.zip',
    ]
    actual_result = build_manager._sort_build_urls_by_revision(
        build_urls, bucket_path, reverse=False)
    self.assertEqual(expected_result, actual_result)

  def test_duplicate_revision(self):
    """Tests that duplicate revision filename results in an exception."""
    bucket_path = ('gs://chromium-browser-libfuzzer/'
                   'linux-release-asan/libfuzzer-linux-release-(35)[0-9]+.zip')
    build_urls = [
        'linux-release-asan/libfuzzer-linux-release-359936.zip',
        'linux-release-asan/libfuzzer-linux-release-359945.zip',
    ]

    with self.assertRaises(errors.BadStateError):
      build_manager._sort_build_urls_by_revision(
          build_urls, bucket_path, reverse=True)

  def test_revision_in_revision(self):
    """Tests that if a revision is a substring of another revision, then it is
    only shown once and not repeated."""
    bucket_path = ('gs://chromium-browser-libfuzzer/'
                   'linux-release-asan/libfuzzer-linux-release-([0-9]+).zip')
    build_urls = [
        'linux-release-asan/libfuzzer-linux-release-359936.zip',
        'linux-release-asan/libfuzzer-linux-release-936.zip',
        'linux-release-asan/libfuzzer-linux-release-359945.zip',
        'linux-release-asan/libfuzzer-linux-release-599.zip',
    ]

    expected_result = [
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-359945.zip',
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-359936.zip',
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-936.zip',
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-599.zip'
    ]
    actual_result = build_manager._sort_build_urls_by_revision(
        build_urls, bucket_path, reverse=True)
    self.assertEqual(expected_result, actual_result)

  def test_duplicate_different_prefix(self):
    """Test that we handle duplicate filenames with different prefixes
    properly."""
    bucket_path = ('gs://chromium-browser-libfuzzer/'
                   'linux-release-asan/libfuzzer-linux-release-([0-9]+).zip')
    build_urls = [
        'linux-release-asan/libfuzzer-linux-release-359936.zip',
        'linux-release-asan/duplicate/libfuzzer-linux-release-359936.zip',
        'linux-release-asan/libfuzzer-linux-release-936.zip',
        'linux-release-asan/duplicate/libfuzzer-linux-release-936.zip',
        'linux-release-asan/duplicate/libfuzzer-linux-release-123.zip',
    ]

    expected_result = [
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-359936.zip',
        'gs://chromium-browser-libfuzzer/'
        'linux-release-asan/libfuzzer-linux-release-936.zip',
    ]
    actual_result = build_manager._sort_build_urls_by_revision(
        build_urls, bucket_path, reverse=True)
    self.assertEqual(expected_result, actual_result)

  def test_bucket_root(self):
    """Tests regular case on bucket root with and without reverse flag set."""
    bucket_path = ('gs://chromium-browser-libfuzzer/'
                   'libfuzzer-linux-release-([0-9]+).zip')
    build_urls = [
        'libfuzzer-linux-release-359936.zip',
        'libfuzzer-linux-release-359950.zip',
        'libfuzzer-linux-release-359945.zip',
        'libfuzzer-linux-release-359953.zip',
    ]
    expected_result = [
        'gs://chromium-browser-libfuzzer/libfuzzer-linux-release-359953.zip',
        'gs://chromium-browser-libfuzzer/libfuzzer-linux-release-359950.zip',
        'gs://chromium-browser-libfuzzer/libfuzzer-linux-release-359945.zip',
        'gs://chromium-browser-libfuzzer/libfuzzer-linux-release-359936.zip'
    ]
    actual_result = build_manager._sort_build_urls_by_revision(
        build_urls, bucket_path, reverse=True)
    self.assertEqual(expected_result, actual_result)

    expected_result = [
        'gs://chromium-browser-libfuzzer/libfuzzer-linux-release-359936.zip',
        'gs://chromium-browser-libfuzzer/libfuzzer-linux-release-359945.zip',
        'gs://chromium-browser-libfuzzer/libfuzzer-linux-release-359950.zip',
        'gs://chromium-browser-libfuzzer/libfuzzer-linux-release-359953.zip',
    ]
    actual_result = build_manager._sort_build_urls_by_revision(
        build_urls, bucket_path, reverse=False)
    self.assertEqual(expected_result, actual_result)


class SplitFuzzTargetsBuildTest(fake_filesystem_unittest.TestCase):
  """Tests for split fuzz target build setup."""

  def setUp(self):
    """Setup for split fuzz targets build test."""
    test_utils.set_up_pyfakefs(self)

    test_helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager.get_build_urls_list',
        'clusterfuzz._internal.build_management.build_manager._make_space',
        'clusterfuzz._internal.build_management.build_manager._make_space_for_build',
        'clusterfuzz._internal.system.shell.clear_temp_directory',
        'clusterfuzz._internal.google_cloud_utils.storage.copy_file_from',
        'clusterfuzz._internal.google_cloud_utils.storage.get_download_file_size',
        'clusterfuzz._internal.google_cloud_utils.storage.list_blobs',
        'clusterfuzz._internal.google_cloud_utils.storage.read_data',
        'clusterfuzz._internal.system.archive.unpack',
        'time.time',
    ])

    test_helpers.patch_environ(self)

    os.environ['BUILDS_DIR'] = '/builds'
    os.environ['FAIL_RETRIES'] = '1'
    os.environ['JOB_NAME'] = 'libfuzzer_job'
    os.environ['UNPACK_ALL_FUZZ_TARGETS_AND_FILES'] = 'True'
    os.environ['FUZZER_DIR'] = os.path.join(os.environ['ROOT_DIR'], 'src',
                                            'clusterfuzz', '_internal', 'bot',
                                            'fuzzers', 'libFuzzer')
    self.fs.add_real_directory(os.environ['FUZZER_DIR'])

    self.mock.list_blobs.return_value = (
        '/subdir/target1/',
        '/subdir/target2/',
        '/subdir/target3/',
        '/subdir/targets.list',
    )
    self.mock.read_data.return_value = (b'target1\ntarget2\ntarget3\n')

    self.target_weights = {
        'target1': 0.0,
        'target2': 1.0,
        'target3': 0.0,
    }
    self.mock.get_download_file_size.return_value = 1
    self.mock.copy_file_from.return_value = True

    self.mock._make_space.return_value = True
    self.mock._make_space_for_build.return_value = True
    self.mock.unpack.return_value = True
    self.mock.time.return_value = 1000.0

    os.environ['FUZZ_TARGET_BUILD_BUCKET_PATH'] = (
        'gs://bucket/subdir/%TARGET%/([0-9]+).zip')

    self.mock.get_build_urls_list.return_value = [
        'gs://bucket/subdir/target2/10.zip',
        'gs://bucket/subdir/target2/2.zip',
        'gs://bucket/subdir/target2/1.zip',
    ]

  def _assert_env_vars(self, target, revision):
    """Assert the expected values of environment variables."""
    self.assertEqual(
        'gs://bucket/subdir/{target}/{revision}.zip'.format(
            target=target, revision=revision), os.environ.get('BUILD_URL'))
    self.assertEqual(str(revision), os.environ['APP_REVISION'])
    self.assertEqual(
        '/builds/bucket_subdir_{target}_'
        '77651789446b3c3a04b9f492ff141f003d437347/revisions'.format(
            target=target),
        os.environ['BUILD_DIR'])
    self.assertEqual('', os.environ['APP_PATH'])

  def test_setup_fuzz(self):
    """Tests setting up a build during fuzzing."""
    os.environ['TASK_NAME'] = 'fuzz'
    self.mock.time.return_value = 1000.0

    build = build_manager.setup_build(target_weights=self.target_weights)
    self.assertIsInstance(build, build_manager.RegularBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1000.0)
    self.assertEqual('target2', os.environ['FUZZ_TARGET'])
    self._assert_env_vars('target2', 10)

    self.assertEqual(1, self.mock.unpack.call_count)
    self.mock.unpack.assert_called_with(
        '/builds/bucket_subdir_target2_77651789446b3c3a04b9f492ff141f003d437347'
        '/revisions/10.zip',
        '/builds/bucket_subdir_target2_77651789446b3c3a04b9f492ff141f003d437347'
        '/revisions',
        file_match_callback=None,
        trusted=True)
    self.assertEqual(3, environment.get_value('FUZZ_TARGET_COUNT'))

  def test_setup_nonfuzz(self):
    """Tests setting up a build during a non-fuzz task."""
    os.environ['FUZZ_TARGET'] = 'target1'
    self.mock.time.return_value = 1000.0

    self.mock.get_build_urls_list.return_value = [
        'gs://bucket/subdir/target1/10.zip',
        'gs://bucket/subdir/target1/8.zip',
    ]

    build = build_manager.setup_build(8)
    self.assertIsInstance(build, build_manager.RegularBuild)
    self.assertEqual(_get_timestamp(build.base_build_dir), 1000.0)
    self.assertEqual('target1', os.environ['FUZZ_TARGET'])
    self._assert_env_vars('target1', 8)

    self.assertEqual(1, self.mock.unpack.call_count)
    self.mock.unpack.assert_called_with(
        '/builds/bucket_subdir_target1_77651789446b3c3a04b9f492ff141f003d437347'
        '/revisions/8.zip',
        '/builds/bucket_subdir_target1_77651789446b3c3a04b9f492ff141f003d437347'
        '/revisions',
        file_match_callback=None,
        trusted=True)
    self.assertIsNone(environment.get_value('FUZZ_TARGET_COUNT'))

  def test_delete(self):
    """Test deleting this build."""
    os.environ['FUZZ_TARGET'] = 'target2'
    build = build_manager.setup_build(10)

    self.assertTrue(
        os.path.isdir('/builds/bucket_subdir_target2_'
                      '77651789446b3c3a04b9f492ff141f003d437347'
                      '/revisions'))
    build.delete()
    self.assertFalse(
        os.path.isdir('/builds/bucket_subdir_target2_'
                      '77651789446b3c3a04b9f492ff141f003d437347'
                      '/revisions'))
    self.assertTrue(
        os.path.isdir('/builds/bucket_subdir_target2_'
                      '77651789446b3c3a04b9f492ff141f003d437347'))

  def test_target_not_built(self):
    """Test a target that's listed in target.list, but not yet built."""
    self.mock.list_blobs.return_value = (
        '/subdir/target1/',
        '/subdir/target3/',
        '/subdir/targets.list',
    )

    targets_list = build_manager._get_targets_list(
        os.environ['FUZZ_TARGET_BUILD_BUCKET_PATH'])
    six.assertCountEqual(self, ['target1', 'target3'], targets_list)


class GetPrimaryBucketPathTest(unittest.TestCase):
  """Tests for get_primary_bucket_path."""

  def setUp(self):
    test_helpers.patch_environ(self)

  def test_release_bucket_path(self):
    """Test primary bucket being a RELEASE_BUILD_BUCKET_PATH."""
    os.environ['RELEASE_BUILD_BUCKET_PATH'] = 'gs://release_build'
    self.assertEqual('gs://release_build',
                     build_manager.get_primary_bucket_path())

  def test_fuzz_target_bucket_path(self):
    """Test primary bucket being a FUZZ_TARGET_BUILD_BUCKET_PATH."""
    os.environ[
        'FUZZ_TARGET_BUILD_BUCKET_PATH'] = 'gs://fuzz_target/%TARGET%/path'
    os.environ['FUZZ_TARGET'] = 'test_target'
    self.assertEqual('gs://fuzz_target/test_target/path',
                     build_manager.get_primary_bucket_path())

  def test_fuzz_target_bucket_path_multi_target(self):
    """Test primary bucket being a FUZZ_TARGET_BUILD_BUCKET_PATH with a multi
    target binary."""
    os.environ[
        'FUZZ_TARGET_BUILD_BUCKET_PATH'] = 'gs://fuzz_target/%TARGET%/path'
    os.environ['FUZZ_TARGET'] = 'test_target@target'
    self.assertEqual('gs://fuzz_target/test_target/path',
                     build_manager.get_primary_bucket_path())

  def test_fuzz_target_bucket_path_no_fuzz_target(self):
    """Test primary bucket being a FUZZ_TARGET_BUILD_BUCKET_PATH with no fuzz
    target defined."""
    os.environ[
        'FUZZ_TARGET_BUILD_BUCKET_PATH'] = 'gs://fuzz_target/%TARGET%/path'
    with self.assertRaises(build_manager.BuildManagerException):
      build_manager.get_primary_bucket_path()

  def test_no_path_defined(self):
    """Test no bucket paths defined."""
    with self.assertRaises(build_manager.BuildManagerException):
      build_manager.get_primary_bucket_path()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Stack analyzer module."""

from clusterfuzz import stacktraces
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.platforms.android import \
    kernel_utils as android_kernel
from clusterfuzz._internal.platforms.linux.lkl import kernel_utils as lkl_kernel
from clusterfuzz._internal.system import environment

MAX_REDZONE_SIZE_FOR_OOMS_AND_HANGS = 64


def linkify_kernel_or_lkl_stacktrace_if_needed(crash_info):
  """Linkify Android Kernel or lkl stacktrace."""
  kernel_prefix = ''
  kernel_hash = ''
  if (environment.is_android_kernel() and
      not environment.is_android_cuttlefish() and
      (crash_info.found_android_kernel_crash or crash_info.is_kasan)):
    kernel_prefix, kernel_hash = \
      android_kernel.get_kernel_prefix_and_full_hash()

  elif (environment.is_lkl_job() and crash_info.is_lkl and
        crash_info.lkl_kernel_build_id):
    kernel_prefix, kernel_hash = \
      lkl_kernel.get_kernel_prefix_and_full_hash(crash_info.lkl_kernel_build_id)

  if kernel_prefix and kernel_hash:
    _linkify_android_kernel_stacktrace(crash_info, kernel_prefix, kernel_hash)


def _linkify_android_kernel_stacktrace(crash_info, kernel_prefix, kernel_hash):
  """Linkify Android Kernel or lkl stacktrace."""
  temp_crash_stacktrace = ''
  for line in crash_info.crash_stacktrace.splitlines():
    temp_crash_stacktrace += android_kernel.get_kernel_stack_frame_link(
        line, kernel_prefix, kernel_hash) + '\n'

  crash_info.crash_stacktrace = temp_crash_stacktrace


def get_crash_data(crash_data,
                   symbolize_flag=True,
                   fuzz_target=None,
                   already_symbolized=False,
                   detect_ooms_and_hangs=None):
  """Get crash parameters from crash data.
  Crash parameters include crash type, address, state and stacktrace.
  If the stacktrace is not already symbolized, we will try to symbolize it
  unless |symbolize| flag is set to False. Symbolized stacktrace will contain
  inline frames, but we do exclude them for purposes of crash state generation
  (helps in testcase deduplication)."""
  # Decide whether to symbolize or not symbolize the input stacktrace.
  # Note that Fuchsia logs are always symbolized.
  if symbolize_flag:
    # Defer imports since stack_symbolizer pulls in a lot of things.
    from clusterfuzz._internal.crash_analysis.stack_parsing import \
        stack_symbolizer
    crash_stacktrace_with_inlines = stack_symbolizer.symbolize_stacktrace(
        crash_data, enable_inline_frames=True)
    crash_stacktrace_without_inlines = stack_symbolizer.symbolize_stacktrace(
        crash_data, enable_inline_frames=False)
  else:
    # We are explicitly indicated to not symbolize using |symbolize_flag|. There
    # is no distinction between inline and non-inline frames for an unsymbolized
    # stacktrace.
    crash_stacktrace_with_inlines = crash_data
    crash_stacktrace_without_inlines = crash_data

  # Additional stack frame ignore regexes.
  custom_stack_frame_ignore_regexes = (
      local_config.ProjectConfig().get('stacktrace.stack_frame_ignore_regexes',
                                       []))

  if environment.get_value('TASK_NAME') == 'analyze':
    detect_v8_runtime_errors = True
  else:
    detect_v8_runtime_errors = environment.get_value('DETECT_V8_RUNTIME_ERRORS',
                                                     False)

  fuzz_target = fuzz_target or environment.get_value('FUZZ_TARGET')
  redzone_size = environment.get_value('REDZONE')
  if detect_ooms_and_hangs is None:
    detect_ooms_and_hangs = (
        environment.get_value('REPORT_OOMS_AND_HANGS') and
        (not redzone_size or
         redzone_size <= MAX_REDZONE_SIZE_FOR_OOMS_AND_HANGS))

  include_ubsan = 'halt_on_error=0' not in environment.get_value(
      'UBSAN_OPTIONS', '')

  stack_parser = stacktraces.StackParser(
      symbolized=symbolize_flag or already_symbolized,
      detect_ooms_and_hangs=detect_ooms_and_hangs,
      detect_v8_runtime_errors=detect_v8_runtime_errors,
      custom_stack_frame_ignore_regexes=custom_stack_frame_ignore_regexes,
      fuzz_target=fuzz_target,
      include_ubsan=include_ubsan)

  result = stack_parser.parse(crash_stacktrace_without_inlines)

  # Use stacktrace with inlines for the result.
  if result.crash_stacktrace:
    result.crash_stacktrace = crash_stacktrace_with_inlines

  # Linkify Android Kernel or lkl stacktrace.
  linkify_kernel_or_lkl_stacktrace_if_needed(result)

  return result



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Authentication helpers."""

import collections

from firebase_admin import auth
from google.cloud import ndb
from googleapiclient.discovery import build
import jwt
import requests

from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from libs import request_cache

User = collections.namedtuple('User', ['email'])


class AuthError(Exception):
  """Auth error."""


def auth_domain():
  """Get the auth domain."""
  domain = local_config.ProjectConfig().get('firebase.auth_domain')
  if domain:
    return domain

  return utils.get_application_id() + '.firebaseapp.com'


def is_current_user_admin():
  """Returns whether or not the current logged in user is an admin."""
  if environment.is_local_development():
    return True

  user = get_current_user()
  if not user:
    return False

  key = ndb.Key(data_types.Admin, user.email)
  return bool(key.get())


@memoize.wrap(memoize.FifoInMemory(1))
def _project_number_from_id(project_id):
  """Get the project number from project ID."""
  resource_manager = build('cloudresourcemanager', 'v1')
  result = resource_manager.projects().get(projectId=project_id).execute()
  if 'projectNumber' not in result:
    raise AuthError('Failed to get project number.')

  return result['projectNumber']


@memoize.wrap(memoize.FifoInMemory(1))
def _get_iap_key(key_id):
  """Retrieves a public key from the list published by Identity-Aware Proxy,
  re-fetching the key file if necessary.
  """
  resp = requests.get('https://www.gstatic.com/iap/verify/public_key')
  if resp.status_code != 200:
    raise AuthError('Unable to fetch IAP keys: {} / {} / {}'.format(
        resp.status_code, resp.headers, resp.text))

  result = resp.json()
  key = result.get(key_id)
  if not key:
    raise AuthError('Key {!r} not found'.format(key_id))

  return key


def _validate_iap_jwt(iap_jwt):
  """Validate JWT assertion."""
  project_id = utils.get_application_id()
  expected_audience = '/projects/{}/apps/{}'.format(
      _project_number_from_id(project_id), project_id)

  try:
    key_id = jwt.get_unverified_header(iap_jwt).get('kid')
    if not key_id:
      raise AuthError('No key ID.')

    key = _get_iap_key(key_id)
    decoded_jwt = jwt.decode(
        iap_jwt,
        key,
        algorithms=['ES256'],
        issuer='https://cloud.google.com/iap',
        audience=expected_audience)
    return decoded_jwt['email']
  except (jwt.exceptions.InvalidTokenError,
          requests.exceptions.RequestException) as e:
    raise AuthError('JWT assertion decode error: ' + str(e))


def get_iap_email(current_request):
  """Get Cloud IAP email."""
  jwt_assertion = current_request.headers.get('X-Goog-IAP-JWT-Assertion')
  if not jwt_assertion:
    return None

  return _validate_iap_jwt(jwt_assertion)


def get_current_user():
  """Get the current logged in user, or None."""
  if environment.is_local_development():
    return User('user@localhost')

  current_request = request_cache.get_current_request()
  if local_config.AuthConfig().get('enable_loas'):
    loas_user = current_request.headers.get('X-AppEngine-LOAS-Peer-Username')
    if loas_user:
      return User(loas_user + '@google.com')

  iap_email = get_iap_email(current_request)
  if iap_email:
    return User(iap_email)

  cache_backing = request_cache.get_cache_backing()
  oauth_email = getattr(cache_backing, '_oauth_email', None)
  if oauth_email:
    return User(oauth_email)

  cached_email = getattr(cache_backing, '_cached_email', None)
  if cached_email:
    return User(cached_email)

  session_cookie = get_session_cookie()
  if not session_cookie:
    return None

  try:
    decoded_claims = decode_claims(get_session_cookie())
  except AuthError:
    logs.log_warn('Invalid session cookie.')
    return None

  if not decoded_claims.get('email_verified'):
    return None

  email = decoded_claims.get('email')
  if not email:
    return None

  # We cache the email for this request if we've validated the user to make
  # subsequent get_current_user() calls fast.
  setattr(cache_backing, '_cached_email', email)
  return User(email)


def create_session_cookie(id_token, expires_in):
  """Create a new session cookie."""
  try:
    return auth.create_session_cookie(id_token, expires_in=expires_in)
  except auth.AuthError:
    raise AuthError('Failed to create session cookie.')


def get_session_cookie():
  """Get the current session cookie."""
  return request_cache.get_current_request().cookies.get('session')


def revoke_session_cookie(session_cookie):
  """Revoke a session cookie."""
  decoded_claims = decode_claims(session_cookie)
  auth.revoke_refresh_tokens(decoded_claims['sub'])


def decode_claims(session_cookie):
  """Decode the claims for the current session cookie."""
  try:
    return auth.verify_session_cookie(session_cookie, check_revoked=True)
  except (ValueError, auth.AuthError):
    raise AuthError('Invalid session cookie.')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Package tests."""
import os
import tempfile
import unittest
import zipfile

from clusterfuzz._internal.system import shell
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from local.butler import package


class IsNodeUpToDateTest(unittest.TestCase):
  """is_nodejs_up_to_date tests."""

  def setUp(self):
    helpers.patch(self, ['local.butler.common.execute'])

  def test_succeed(self):
    """Test when succeed."""
    self.mock.execute.return_value = (0, b'v5.3.11')
    self.assertTrue(package._is_nodejs_up_to_date())  # pylint: disable=protected-access
    self.mock.execute.assert_called_once_with('node -v')

  def test_fail_return_code(self):
    """Test return code is non-zero."""
    self.mock.execute.return_value = (1, b'v5.3.11')
    self.assertFalse(package._is_nodejs_up_to_date())  # pylint: disable=protected-access
    self.mock.execute.assert_called_once_with('node -v')

  def test_fail_parse(self):
    """Test when output cannot be parse."""
    self.mock.execute.return_value = (0, b'sdafsadf')
    self.assertFalse(package._is_nodejs_up_to_date())  # pylint: disable=protected-access
    self.mock.execute.assert_called_once_with('node -v')

  def test_fail_version(self):
    """Test version is less than 4."""
    self.mock.execute.return_value = (0, b'v3.2.1')
    self.assertFalse(package._is_nodejs_up_to_date())  # pylint: disable=protected-access
    self.mock.execute.assert_called_once_with('node -v')


@test_utils.slow
@test_utils.integration
class PackageTest(unittest.TestCase):
  """Package tests."""

  def setUp(self):
    if os.getenv('PARALLEL_TESTS'):
      self.skipTest('Package testing is disabled when running in parallel.')

    self.temp_directory = tempfile.mkdtemp()
    self.zip_directory = os.path.join(self.temp_directory, 'packages')
    self.manifest_filename = os.path.join(self.temp_directory, 'test.manifest')

    helpers.patch(self, ['local.butler.common.is_git_dirty'])
    self.mock.is_git_dirty.return_value = False

  def tearDown(self):
    shell.remove_directory(self.temp_directory)

  def test_package(self):
    """Test package."""
    package.package(
        'revision',
        self.zip_directory,
        self.manifest_filename,
        platform_name='linux')

    zip_filename = os.path.join(self.zip_directory, 'linux.zip')
    with zipfile.ZipFile(zip_filename, 'r') as f:
      files = f.namelist()

      manifest_file = os.path.join('clusterfuzz', 'src', 'appengine',
                                   'resources', 'clusterfuzz-source.manifest')

      self.assertIn(manifest_file, files)

      self.assertEqual(b'revision\n', f.read(manifest_file))

      self.assertIn(
          os.path.join('clusterfuzz', 'src', 'third_party', 'googleapiclient',
                       '__init__.py'), files)
      self.assertIn(
          os.path.join('clusterfuzz', 'src', 'appengine', 'config', 'gae',
                       'auth.yaml'), files)
      self.assertNotIn(
          os.path.join('clusterfuzz', 'configs', 'test', 'gae', 'auth.yaml'),
          files)
      self.assertNotIn(
          os.path.join('clusterfuzz', 'local', '__init__.py'), files)
      self.assertNotIn(
          os.path.join('clusterfuzz', 'src', 'appengine', 'app.yaml'), files)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Build setup host (client)."""

from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz._internal.system import environment

from . import host


def _clear_env():
  """Clear build env vars."""
  environment.remove_key('APP_PATH')
  environment.remove_key('APP_REVISION')
  environment.remove_key('APP_PATH_DEBUG')
  environment.remove_key('APP_DIR')
  environment.remove_key('BUILD_DIR')
  environment.remove_key('BUILD_URL')
  environment.remove_key('FUZZ_TARGET')


def _handle_response(build, response):
  """Handle build setup response."""
  if not response.result:
    _clear_env()
    return False

  _update_env_from_response(response)

  if not environment.get_value('APP_PATH'):
    fuzzer_directory = environment.get_value('FUZZER_DIR')
    if fuzzer_directory:
      build_manager.set_environment_vars([fuzzer_directory])

  environment.set_value('APP_REVISION', build.revision)
  return True


def _update_env_from_response(response):
  """Update environment variables from response."""
  environment.set_value('APP_PATH', response.app_path)
  environment.set_value('APP_PATH_DEBUG', response.app_path_debug)
  environment.set_value('APP_DIR', response.app_dir)
  environment.set_value('BUILD_DIR', response.build_dir)
  environment.set_value('BUILD_URL', response.build_url)
  environment.set_value('FUZZ_TARGET', response.fuzz_target)
  environment.set_value('FUZZ_TARGET_COUNT', response.fuzz_target_count)


class RemoteRegularBuild(build_manager.RegularBuild):
  """Remote regular build."""

  def setup(self):
    request = untrusted_runner_pb2.SetupRegularBuildRequest(
        base_build_dir=self.base_build_dir,
        revision=self.revision,
        build_url=self.build_url,
        build_prefix=self.build_prefix)
    if self.target_weights:
      request.target_weights.update(self.target_weights)

    return _handle_response(self, host.stub().SetupRegularBuild(request))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for updating from trunk. In other words, updating the stacktrace."""
from clusterfuzz._internal.base import tasks
from handlers import base_handler
from handlers.testcase_detail import show
from libs import handler
from libs import helpers


def update(testcase):
  """Update from trunk."""
  testcase.last_tested_crash_stacktrace = 'Pending'
  testcase.put()

  tasks.add_task(
      'variant',
      testcase.key.id(),
      testcase.job_type,
      queue=tasks.queue_for_testcase(testcase))

  helpers.log(
      'Marked testcase %s for last tested stacktrace update' %
      testcase.key.id(), helpers.MODIFY_OPERATION)


class Handler(base_handler.Handler):
  """Handler that updates from trunk."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.require_csrf_token
  @handler.check_testcase_access
  def post(self, testcase):
    """Update from trunk."""
    update(testcase)
    return self.render_json(show.get_testcase_detail(testcase))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Minimizer used for additional reduction on javascript test cases."""

from clusterfuzz._internal.bot.tokenizer.antlr_tokenizer import AntlrTokenizer
from clusterfuzz._internal.bot.tokenizer.grammars.JavaScriptLexer import \
    JavaScriptLexer

from . import delta_minimizer
from . import errors
from . import minimizer
from . import utils


def step_back_while(cur_index, condition):
  """Helper function. Decreases index from cur while condition is satisfied."""
  while cur_index >= 0 and condition(cur_index):
    cur_index -= 1
  return cur_index


class JSMinimizer(minimizer.Minimizer):
  """Intended as a second-pass minimizer to remove unneeded tokens from JS."""

  def _execute(self, data):
    testcase = minimizer.Testcase(data, self)
    if not self.validate_tokenizer(data, testcase):
      raise errors.TokenizationFailureError('JS Minimizer')

    brace_stack = []
    paren_stack = []

    for index, token in enumerate(testcase.tokens):
      if token == '{':
        brace_stack.append(index)

      elif token == '}' and brace_stack:

        # Two hypotheses for tokens grouped by curly braces:
        # 1) Remove from start of line to open brace and the closing brace.
        #    e.g.: if (statement_that_evaluates_to_true) { crash() } -> crash()
        open_brace_index = brace_stack.pop()

        # Find the first non-empty token prior to the starting brackets.
        token_before_bracket = step_back_while(
            open_brace_index - 1, (lambda x: not testcase.tokens[x].strip()))

        # If that token is a close paren, we need to grab everything else too.
        # Do this to grab the whole paren so we don't create a syntax error by
        # removing only part of a paren.
        if testcase.tokens[token_before_bracket] == ')':
          # Find everything in the paren.
          token_before_bracket = step_back_while(
              token_before_bracket, (lambda x: testcase.tokens[x] != '('))

          # Get the token before the paren.
          token_before_bracket -= 1
          token_before_bracket = step_back_while(
              token_before_bracket, (lambda x: not testcase.tokens[x].strip()))

        # Walk back to the start of that line as well to get if/else and funcs.
        # Do this after paren to manage situations where there are newlines in
        # the parens.
        token_before_bracket = step_back_while(
            token_before_bracket, (lambda x: testcase.tokens[x] != '\n'))

        token_before_bracket += 1

        hypothesis = list(range(token_before_bracket,
                                open_brace_index + 1)) + [index]

        testcase.prepare_test(hypothesis)

        # 2) Remove previous tokens and from the closing brace to the next one.
        #    e.g.: try { crash() } catch(e) {} -> crash().
        future_index = len(testcase.tokens)
        open_count = 0
        for future_index in range(index + 1, len(testcase.tokens)):
          if testcase.tokens[future_index] == '{':
            open_count += 1
          if testcase.tokens[future_index] == '}':
            open_count -= 1
            # Make sure to grab entire outer brace if there are inner braces.
            if not open_count:
              break
        if future_index != len(testcase.tokens):
          lookahead_hypothesis = hypothesis + list(
              range(index + 1, future_index + 1))

          testcase.prepare_test(lookahead_hypothesis)

      elif token == '(':
        paren_stack.append(index)

      elif token == ')' and paren_stack:
        # Three hypotheses for tokens grouped by parentheses:
        # 1) Remove the parentheses and the previous token.
        #    e.g.: assertTrue(crash()); -> crash()
        previous_end = paren_stack.pop()
        if previous_end > 0:
          hypothesis = [previous_end - 1, previous_end, index]
          testcase.prepare_test(hypothesis)

        # 2) Remove everything between the parentheses.
        #    e.g. crash(junk, more_junk) -> crash()
        if index - previous_end > 1:
          hypothesis = list(range(previous_end + 1, index))
          testcase.prepare_test(hypothesis)

        # 3) Like 1, but to start of line instead of previous token.
        #    e.g.: leftover_junk = (function() {
        #          });

        # Find the beginning of the line
        token_before_paren = previous_end
        token_before_paren = step_back_while(
            previous_end, (lambda x: testcase.tokens[x] != '\n'))
        token_before_paren += 1

        hypothesis = list(range(token_before_paren, previous_end + 1)) + [index]
        testcase.prepare_test(hypothesis)

        # 4) Like 3, but also from the closing brace to the next one.
        #    e.g.: (function(global) { })(this);
        future_index = len(testcase.tokens)
        for future_index in range(index + 1, len(testcase.tokens)):
          if testcase.tokens[future_index] == ')':
            break
        if future_index != len(testcase.tokens):
          lookahead_hypothesis = list(
              range(token_before_paren, future_index + 1))
          testcase.prepare_test(lookahead_hypothesis)

      elif token == ',':
        # Two hypotheses for commas:
        # 1) Remove comma and left-hand-side.
        #    e.g.: f(whatever, crash()) -> f(crash())
        if index > 0:
          hypothesis = [index - 1, index]
          testcase.prepare_test(hypothesis)

        # 2) Remove comma and right-hand-side.
        #    e.g.: f(crash(), whatever) -> f(crash())

        # Find the next non whitespace token after the comma.
        hypothesis = [index]
        for right_token_index in range(index + 1, len(testcase.tokens)):
          hypothesis.append(right_token_index)
          if testcase.tokens[right_token_index].strip():
            testcase.prepare_test(hypothesis)
            break

    testcase.process()
    return testcase

  @staticmethod
  def run(data,
          thread_count=minimizer.DEFAULT_THREAD_COUNT,
          file_extension='.js'):
    """Attempt to minimize a javascript test case."""
    line_minimizer = delta_minimizer.DeltaMinimizer(
        utils.test, max_threads=thread_count, file_extension=file_extension)

    js_tokenizer = AntlrTokenizer(JavaScriptLexer)

    js_minimizer = JSMinimizer(
        utils.test,
        max_threads=thread_count,
        tokenizer=js_tokenizer.tokenize,
        token_combiner=js_tokenizer.combine,
        file_extension=file_extension)

    result = line_minimizer.minimize(data)
    result = js_minimizer.minimize(result)
    result = js_minimizer.minimize(result)
    result = line_minimizer.minimize(result)

    return result



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Help."""

from clusterfuzz._internal.config import db_config
from handlers import base_handler

DEFAULT_DOCUMENTATION_URL = 'https://google.github.io/clusterfuzz/'
DEFAULT_BUG_REPORT_URL = 'https://github.com/google/clusterfuzz/issues'


class DocumentationHandler(base_handler.Handler):
  """Redirect to documentation."""

  def get(self):
    """Get the HTML page."""
    documentation_url = db_config.get_value('documentation_url')
    if not documentation_url:
      documentation_url = DEFAULT_DOCUMENTATION_URL

    return self.redirect(documentation_url)


class ReportBugHandler(base_handler.Handler):
  """Redirect to issue tracker for reporting bug."""

  def get(self):
    """Get the HTML page."""
    bug_report_url = db_config.get_value('bug_report_url')
    if not bug_report_url:
      bug_report_url = DEFAULT_BUG_REPORT_URL

    return self.redirect(bug_report_url)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for the crash stats page."""

import json

from flask import request

from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.metrics import crash_stats as crash_stats_common
from handlers import base_handler
from libs import crash_access
from libs import crash_stats
from libs import filters
from libs import handler
from libs import helpers

PAGE_SIZE = 10
DEFAULT_DAYS_FOR_BY_HOURS = 3
DEFAULT_DAYS_FOR_BY_DAYS = 7


class FuzzerFilter(filters.Filter):
  """Filter for fuzzer."""

  def add(self, query, params):
    """Set query according to fuzzer param."""
    value = params.get('fuzzer', '')
    if filters.is_empty(value):
      return

    if data_handler.is_fuzzing_engine(value):
      query.filter('parent_fuzzer_name', value)
    else:
      query.filter('fuzzer_name', value)


class PlatformFilter(filters.Filter):
  """Filter for platform."""

  def add(self, query, params):
    """Set query according to platform param."""
    value = params.get('platform', '')
    if filters.is_empty(value):
      return

    if value == 'android':
      query.filter('parent_platform', value)
    else:
      query.filter('platform', value)


class TimeFilter(filters.Filter):
  """Filter for start and end hour."""

  def add(self, query, params):
    """Set query according to end and hours params."""
    if 'end' in params:
      end = helpers.cast(params['end'], int, "'end' must be an integer")
    else:
      end = crash_stats_common.get_max_hour()

    block = params.get('block', 'day')

    if 'days' in params:
      days = helpers.cast(params['days'], int, "'days' must be an integer")
    else:
      days = (
          DEFAULT_DAYS_FOR_BY_HOURS
          if block == 'hour' else DEFAULT_DAYS_FOR_BY_DAYS)

    params['end'] = str(end)
    params['days'] = str(days)
    params['block'] = str(block)

    query.set_time_params(end, days, block)


class KeywordFilter(filters.Filter):
  """Filter for keyword."""

  def add(self, query, params):
    """Set query according to search param."""
    value = params.get('q', '')
    if filters.is_empty(value):
      return

    for keyword in value.split(' '):
      query.raw_filter(
          '(LOWER(crash_state) LIKE %s OR LOWER(crash_type) LIKE %s)' %
          (json.dumps('%%%s%%' % keyword.lower()),
           json.dumps('%%%s%%' % keyword.lower())))


GROUP_FILTERS = [
    filters.Boolean('is_new', 'new'),
]

FILTERS = [
    TimeFilter(),
    filters.Boolean('security_flag', 'security'),
    filters.Boolean('reproducible_flag', 'reproducible'),
    filters.String('job_type', 'job'),
    filters.String('project', 'project'),
    FuzzerFilter(),
    PlatformFilter(),
    KeywordFilter(),
]


def query_testcase(project_name, crash_type, crash_state, security_flag,
                   is_open):
  """Start a query for an associated testcase."""
  return data_types.Testcase.query(
      data_types.Testcase.project_name == project_name,
      data_types.Testcase.crash_type == crash_type,
      data_types.Testcase.crash_state == crash_state,
      data_types.Testcase.security_flag == security_flag,
      data_types.Testcase.open == is_open,
      ndb_utils.is_false(data_types.Testcase.is_a_duplicate_flag)).order(
          -data_types.Testcase.timestamp).iter(
              limit=1, projection=[
                  'bug_information',
                  'group_bug_information',
              ])


def attach_testcases(rows):
  """Attach testcase to each crash."""
  testcases = {}
  for index, row in enumerate(rows):
    testcases[index] = {
        'open_testcase':
            query_testcase(
                project_name=row['projectName'],
                crash_type=row['crashType'],
                crash_state=row['crashState'],
                security_flag=row['isSecurity'],
                is_open=True),
        'closed_testcase':
            query_testcase(
                project_name=row['projectName'],
                crash_type=row['crashType'],
                crash_state=row['crashState'],
                security_flag=row['isSecurity'],
                is_open=False)
    }

  for index, row in enumerate(rows):
    testcase = (list(testcases[index]['open_testcase']) or
                list(testcases[index]['closed_testcase']) or [None])[0]
    if testcase:
      testcase = {
          'id': testcase.key.id(),
          'issueNumber': testcase.bug_information,
          'groupIssueNumber': testcase.group_bug_information
      }
    row['testcase'] = testcase


def get_result():
  """Get the result for the crash stats page."""
  params = dict(request.iterparams())
  page = helpers.cast(request.get('page') or 1, int, "'page' is not an int.")
  group_by = params.get('group', 'platform')
  params['group'] = group_by
  sort_by = params.get('sort', 'total_count')
  params['sort'] = sort_by
  params['number'] = params.get('number', 'count')

  # Conditions for individual records.
  query = crash_stats.Query()
  query.group_by = group_by
  query.sort_by = sort_by
  crash_access.add_scope(query, params, 'security_flag', 'job_type',
                         'fuzzer_name')
  filters.add(query, params, FILTERS)

  # Conditions after grouping.
  group_query = crash_stats.Query()
  filters.add(group_query, params, GROUP_FILTERS)

  try:
    total_count, rows = crash_stats.get(
        query=query,
        group_query=group_query,
        offset=(page - 1) * PAGE_SIZE,
        limit=PAGE_SIZE)
  except ValueError:
    raise helpers.EarlyExitException('Invalid filters', 400)

  attach_testcases(rows)

  helpers.log('CrashStats', helpers.VIEW_OPERATION)

  result = {
      'totalPages': (total_count // PAGE_SIZE) + 1,
      'page': page,
      'pageSize': PAGE_SIZE,
      'items': rows,
      'totalCount': total_count
  }
  return result, params


def get_all_platforms():
  """Get all platforms including parent platform."""
  items = data_types.Testcase.query(
      projection=[data_types.Testcase.platform], distinct=True)

  return sorted(
      list(
          set([item.platform.lower() for item in items if item.platform] +
              ['android'])))


class Handler(base_handler.Handler):
  """Handler that gets the crash stats when user first lands on the page."""

  @handler.get(handler.HTML)
  @handler.unsupported_on_local_server
  def get(self):
    """Get and render the crash stats in HTML."""
    result, params = get_result()
    field_values = {
        'fuzzers':
            data_handler.get_all_fuzzer_names_including_children(
                include_parents=True),
        'jobs':
            data_handler.get_all_job_type_names(),
        'platforms':
            get_all_platforms(),
        'projects':
            data_handler.get_all_project_names(),
        'minHour':
            crash_stats_common.get_min_hour(),
        'maxHour':
            crash_stats_common.get_max_hour()
    }
    return self.render('crash-stats.html', {
        'result': result,
        'fieldValues': field_values,
        'params': params
    })


class JsonHandler(base_handler.Handler):
  """Handler that gets the crash stats when user interacts with the page."""

  @handler.post(handler.JSON, handler.JSON)
  def post(self):
    """Get and render the crash stats in JSON."""
    result, _ = get_result()
    return self.render_json(result)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The module chooses the leader given a group of testcases. It needs to be here
  because it is used by group_task.py and attribute_builder.py."""

import itertools

import six


def find_index(items, condition_fn):
  """Return the index of the first item whose condition_fn is True."""
  for index, item in enumerate(items):
    if condition_fn(item):
      return index
  return None


def is_reproducible(item):
  """Return True if the testcase is reproducible by checking the
    one_time_crasher_flag."""
  return not item.one_time_crasher_flag


def has_issue(item):
  """Return True if the testcase has an issue."""
  return bool(item.issue_id)


def is_reproducible_and_has_issue(item):
  """Return True if the testcase is reproducible and has an issue."""
  return is_reproducible(item) and has_issue(item)


def choose(testcase_map):
  """Choose one leader for each group. We choose the highest quality testcase to
    be the leader.

    Args:
      testcase_map: a dict of (testcase_id, testcase). A dict contains testcases
          from multiple groups.
  """

  def _key_func(testcase):
    return testcase.group_id

  testcases = sorted([v for _, v in six.iteritems(testcase_map)], key=_key_func)
  for group_id, items in itertools.groupby(testcases, _key_func):
    if group_id == 0:  # group_id=0 means there's no group.
      continue

    items = sorted(items, reverse=True, key=lambda t: t.timestamp)

    for item in items:
      item.is_leader = False

    leader_index = find_index(items, is_reproducible_and_has_issue)
    if leader_index is None:
      leader_index = find_index(items, has_issue)
    if leader_index is None:
      leader_index = find_index(items, is_reproducible)
    if leader_index is None:
      leader_index = 0

    items[leader_index].is_leader = True



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GCE metadata."""

import socket

import requests

from clusterfuzz._internal.base import retry
from clusterfuzz._internal.system import environment

_METADATA_SERVER = 'metadata.google.internal'

_RETRIES = 3
_DELAY = 1


@retry.wrap(
    retries=_RETRIES,
    delay=_DELAY,
    function='python.google_cloud_utils.compute_metadata.get')
def get(path):
  """Get GCE metadata value."""
  attribute_url = (
      'http://{}/computeMetadata/v1/'.format(_METADATA_SERVER) + path)
  headers = {'Metadata-Flavor': 'Google'}
  operations_timeout = environment.get_value('URL_BLOCKING_OPERATIONS_TIMEOUT')

  response = requests.get(
      attribute_url, headers=headers, timeout=operations_timeout)
  response.raise_for_status()
  return response.text


def is_gce():
  """Return whether or not we're on GCE."""
  try:
    sock = socket.create_connection((_METADATA_SERVER, 80))
    sock.close()
  except Exception:
    return False

  return True
